import unittest
from clusto.test import testbase

import clusto
from clusto.schema import *
from clusto.drivers.base import *
from clusto.drivers import BasicDatacenter
from sqlalchemy.exc import InvalidRequestError

class TestClustoCounter(testbase.ClustoTestBase):

    def testCounterDefault(self):


        e = Entity('e1')
        c = Counter(e, 'key1')

        self.assertEqual(c.value, 0)

        d = Counter(e, 'key2', start=10)

        self.assertEqual(d.value, 10)

    def testCounterIncrement(self):
    
        e = Entity('e1')
        c = Counter(e, 'key1')

                         
        c.next()
        c.next()
        self.assertEqual(c.value,2)

    def testGetCounter(self):

        e = Entity('e1')

        c = Counter.get(e, 'key1')

        c.next()
        self.assertEqual(c.value, 1)


        d = Counter.get(e, 'key1', default=100)
        d.next()
        self.assertEqual(d.value, 2)

        f = Counter.get(e, 'key2', default=20)
        self.assertEqual(f.value, 20)




"""Driver class

A Driver provides an interface to an Entity and its Attributes.
"""

import re
import itertools
import logging

import clusto
from clusto.schema import *
from clusto.exceptions import *

from clusto.drivers.base.clustodriver import *

try:
    import simplejson as json
except:
    import json

class Driver(object):
    """Base Driver.

    The Driver class provides a proxy interface for managing and Entity and
    its Attributes. It provides many helper functions includeing attribute
    setters and accessors, attribute querying, and a handful of conventions.

    Every driver defines a _clusto_type and a _driver_name member variable.
    Upon creation these become the type and driver for the Entity and provides
    a mechanism for choosing the correct driver for a given Entity.

    A Driver can be created by passing either the name (a string) for a new
    Entity you'd like to create, an already instantiated Entity object, or a
    Driver object (which has already been instantiated and is managing an
    Entity).

    If a _properties member dictionary is defined they will be treated as
    default values for the given Entity attributes as well as exposed via a
    simpler mydriver.key access pattern.  So for:

    >>> class MyDriver(Driver):
    >>>    ...
    >>>    _properties = {'propA': 10, 'propB': "default1"}
    >>>    ...

    >>> d = MyDriver('foo')
    >>> d.propA == 10
    True

    >>> d.propB == "default1"
    True

    Only properties with non-None default values are set in the clusto db at
    initial instantiation time (when creating a brand new entity).

    >>> d.propA = 54
    >>> d.propA == 54
    True

    Several conventions are also exposed via the Driver interface.

    """

    __metaclass__ = ClustoDriver

    _clusto_type = "generic"
    _driver_name = "entity"

    _properties = dict()


    @property
    def type(self):
        return self.entity.type

    @property
    def driver(self):
        return self.entity.driver

    def __new__(cls, name_driver_entity=None, *args, **kwargs):

        if isinstance(name_driver_entity, Driver):
            return name_driver_entity
        else:
            return object.__new__(cls)

    def __init__(self, name_driver_entity, **kwargs):

        if not isinstance(name_driver_entity, (str, unicode, Entity, Driver)):
            raise TypeError("First argument must be a string, "
                            "Driver, or Entity.")

        if isinstance(name_driver_entity, Driver):
            return

        if isinstance(name_driver_entity, Entity):

            self.entity = name_driver_entity
            self._choose_best_driver()
            return
        elif isinstance(name_driver_entity, (str, unicode)):

            try:
                existing = clusto.get_by_name(name_driver_entity)
            except LookupError, x:
                existing = None

            if existing:
                raise NameException("Driver with the name %s already exists."
                                    % (name_driver_entity))


            self.entity = Entity(name_driver_entity,
                                 driver=self._driver_name,
                                 clustotype=self._clusto_type)


        else:
            raise TypeError("Could not create driver from given arguments.")

        for key, val in self._properties.iteritems():
            if key in kwargs:
                val = kwargs[key]
            if val is None:
                continue
            setattr(self, key, val)


    def __eq__(self, other):

        if isinstance(other, Entity):
            return self.entity.name == other.name
        elif isinstance(other, Driver):
            return self.entity.name == other.entity.name
        else:
            return False

    def __repr__(self):

        s = "%s(name=%s, type=%s, driver=%s)"

        return s % (self.__class__.__name__, self.entity.name,
                    self.entity.type, self.entity.driver)

    def __cmp__(self, other):

        if hasattr(other, 'name'):
            return cmp(self.name, other.name)
        elif other is None:
            return 1
        else:
            raise TypeError("Cannot compare %s with %s", type(self), type(other))

    def __hash__(self):
        return hash(self.entity.name)

    def __contains__(self, other):
        return self.has_attr(key="_contains", value=other)

    def __iter__(self):
        return iter(self.contents())

    def _choose_best_driver(self):
        """
        Examine the attributes of our entity and set the best driver class and
        mixins.
        """

        self.__class__ = DRIVERLIST[self.entity.driver]



    name = property(lambda x: x.entity.name)


    def _check_attr_name(self, key):
        """
        check to make sure the key does not contain invalid characters
        raise NameException if fail.
        """

        if not isinstance(key, basestring):
            raise TypeError("An attribute name must be a string.")

        if not re.match('^[A-Za-z_]+[0-9A-Za-z_-]*$', key):

            raise NameException("Attribute name %s is invalid. "
                                "Attribute names may not contain periods or "
                                "comas." % key)


    def __getattr__(self, name):
        if name in self._properties:
            attr = self.attr_query(name, subkey='property')
            if not attr:
                return self._properties[name]
            else:
                return attr[0].value
        else:
            raise AttributeError("Attribute %s does not exist." % name)


    def __setattr__(self, name, value):

        if name in self._properties:
            self.set_attr(name, value, subkey='property')
        else:
            object.__setattr__(self, name, value)

    @classmethod
    def ensure_driver(self, obj, msg=None):
        """Ensure that the given argument is a Driver.

        If the object is an Entity it will be turned into a Driver and then
        returned.  If it's a Driver it will be returned unaffected.  Otherwise
        a TypeError is raised with either a generic or given message.
        """

        if isinstance(obj, Entity):
            d = Driver(obj)
        elif isinstance(obj, Driver):
            d = obj
        else:
            if not msg:
                msg = "Not a Driver."
            raise TypeError(msg)

        return d

    @classmethod
    def do_attr_query(cls, key=(), value=(), number=(), start_timestamp=(), end_timestamp=(),
                    subkey=(), ignore_hidden=True, sort_by_keys=False,
                    glob=False, count=False, querybase=None, return_query=False,
                    entity=None):
        """Does queries against all Attributes using the DB."""

        clusto.flush()
        if querybase:
            query = querybase
        else:
            query = Attribute.query()

        ### This is bunk, gotta fix it
        if isinstance(cls, Driver):
            query = query.filter(and_(Attribute.entity_id==Entity.entity_id,
                                      Entity.driver == cls._driver_name,
                                      Entity.type == cls._clusto_type))

        if start_timestamp != () and end_timestamp != ():
            query = query.filter(and_((Attribute.datetime_value >= start_timestamp)))
            query = query.filter(and_(Attribute.datetime_value <= end_timestamp))

        if entity:
            query = query.filter_by(entity_id=entity.entity_id)

        if key is not ():
            key = unicode(key)
        
            if glob:
                query = query.filter(Attribute.key.like(key.replace('*', '%')))
            else:
                query = query.filter_by(key=key)

        if subkey is not ():
            if subkey is not None:
                subkey = unicode(subkey)
        
            if glob and subkey:
                query = query.filter(Attribute.subkey.like(subkey.replace('*', '%')))
            else:
                query = query.filter_by(subkey=subkey)

        if value is not ():
            typename = Attribute.get_type(value)

            if typename == 'relation':
                if isinstance(value, Driver):
                    value = value.entity.entity_id
                query = query.filter_by(relation_id=value)
            else:
                if typename == 'json':
                    typename = 'string'
                    value = json.dumps(value)
                
                if typename == 'string':
                    value = unicode(value)
                
                query = query.filter_by(**{typename+'_value':value})

        if number is not ():
            if isinstance(number, bool) or number is None:
                if number == True:
                    query = query.filter(Attribute.number != None)
                else:
                    query = query.filter(Attribute.number == None)
            elif isinstance(number, (int, long)):
                query = query.filter_by(number=number)

            else:
                raise TypeError("number must be either a boolean or an integer.")

        if ignore_hidden and ((key and not key.startswith('_')) or key is ()):
            query = query.filter(not_(Attribute.key.like(u'@_%', escape=u'@')))

        if sort_by_keys:
            query = query.order_by(Attribute.key)

        if count:
            return query.count()

        if return_query:
            return query

        return query.all()

    def attr_query(self, *args, **kwargs):
        """Queries all attributes of *this* entity using the DB."""

        kwargs['entity'] = self.entity

        return self.do_attr_query(*args, **kwargs)

    @classmethod
    def attr_filter(cls, attrlist, key=(), value=(), number=(),
                   subkey=(), ignore_hidden=True,
                   sort_by_keys=True,
                   regex=False,
                   clusto_types=None,
                   clusto_drivers=None,
                   ):
        """Filter attribute lists. (Uses generator comprehension)

        Given a list of Attributes filter them based on exact matches of key,
        number, subkey, value.

        There are some special cases:

        if number is True then the number variable must be non-null. if
        number is False then the number variable must be null.

        if ignore_hidden is True (the default) then filter out keys that begin
        with an underscore, if false don't filter out such keys.  If you
        specify a key that begins with an underscore as one of the arguments
        then ignore_hidden is assumed to be False.

        if sort_by_keys is True then attributes are returned sorted by keys,
        otherwise their order is undefined.

        if regex is True then treat the key, subkey, and value query
        parameters as regular expressions.

        clusto_types is a list of types that the entities referenced by
        relation attributes must match.

        clusto_drivers is a list of drivers that the entities referenced by
        relation attributes must match.
        """


        result = attrlist

        def subfilter(attrs, val, name):

            if regex:
                testregex = re.compile(val)
                result = (attr for attr in attrs
                          if testregex.match(getattr(attr, name)))

            else:
                result = (attr for attr in attrs
                          if getattr(attr, name) == val)


            return result

        parts = ((key, 'key'), (subkey, 'subkey'), (value, 'value'))
        argattr = ((val,name) for val,name in parts if val is not ())

        for v, n in argattr:
            result = subfilter(result, v, n)


        if number is not ():
            if isinstance(number, bool) or number is None:
                if number:
                    result = (attr for attr in result if attr.number is not None)
                else:
                    result = (attr for attr in result if attr.number is None)

            elif isinstance(number, (int, long)):
                result = (attr for attr in result if attr.number == number)

            else:
                raise TypeError("number must be either a boolean or an integer.")


        if value:
            result = (attr for attr in result if attr.value == value)


        if key and key.startswith('_'):
            ignore_hidden = False

        if ignore_hidden:
            result = (attr for attr in result if not attr.key.startswith('_'))

        if clusto_drivers:
            cdl = [clusto.get_driver_name(n) for n in clusto_drivers]
            result = (attr for attr in result if attr.is_relation and attr.value.entity.driver in cdl)

        if clusto_types:
            ctl = [clusto.get_type_name(n) for n in clusto_types]
            result = (attr for attr in result if attr.is_relation and attr.value.entity.type in ctl)

        if sort_by_keys:
            result = sorted(result)


        return list(result)

    def _itemize_attrs(self, attrlist):
        return [(x.keytuple, x.value) for x in attrlist]

    def attrs(self, *args, **kwargs):
        """Return attributes for this entity.

        (filters whole attribute list as opposed to querying the db directly)
        """

        merge_container_attrs = kwargs.pop('merge_container_attrs', False)
        ignore_memcache = kwargs.pop('ignore_memcache', False)

        if clusto.SESSION.memcache and not ignore_memcache:
            logging.debug('Pulling info from memcache when possible for %s' % self.name)
            k = None
            if 'key' in kwargs:
                k = kwargs['key']
            else:
                if len(args) > 1:
                    k = args[0]
            if k:
#               This is hackish, need to find another way to know if we should cache things or not
                if not k.startswith('_'):
                    if 'subkey' in kwargs and kwargs['subkey'] is not None:
                        memcache_key = str('%s.%s.%s' % (self.name, k, kwargs['subkey']))
                    else:
                        memcache_key = str('%s.%s' % (self.name, k))
                    logging.debug('memcache key: %s' % memcache_key)
                    attrs = clusto.SESSION.memcache.get(memcache_key)
                    if not attrs:
                        attrs = self.attr_filter(self.entity.attrs, *args, **kwargs)
                        if attrs:
                            clusto.SESSION.memcache.set(memcache_key, attrs)
                else:
                    attrs = self.attr_filter(self.entity.attrs, *args, **kwargs)
            else:
                logging.debug('We cannot cache attrs without a key at least')
                attrs = self.attr_filter(self.entity.attrs, *args, **kwargs)
        else:
            attrs = self.attr_filter(self.entity.attrs, *args, **kwargs)

        if merge_container_attrs:
            kwargs['merge_container_attrs'] = merge_container_attrs
            kwargs['ignore_memcache'] = ignore_memcache
            for parent in self.parents():
                for a in parent.attrs(*args,  **kwargs):
                    if a not in attrs:
                        attrs.append(a)

        return attrs

    def attr_values(self, *args, **kwargs):
        """Return the values of the attributes that match the given arguments"""

        return [k.value for k in self.attrs(*args, **kwargs)]

    def attr_value(self, *args, **kwargs):
        """Return a single value for the given arguments or the default if none exist

        extra parameters:
          default - the default value to return if none exist
        """

        if 'default' in kwargs:
            default = kwargs.pop('default')
        else:
            default = None

        vals = self.attr_values(*args, **kwargs)

        if vals:
            if kwargs.get('merge_container_attrs', False):
                return vals[0]
            if len(vals) != 1:
                raise DriverException("args match more than one value")
            return vals[0]
        else:
            return default

    def references(self, *args, **kwargs):
        """Return the references to this Thing. The references are attributes.

        Accepts the same arguments as attrs().

        The semantics of clusto_types and clusto_drivers changes to match the
        clusto_type or clusto_driver of the Entity that owns the attribute as
        opposed to the Entity the attribute refers to.
        """


        clusto_drivers = kwargs.pop('clusto_drivers', None)

        clusto_types = kwargs.pop('clusto_types', None)

        result = self.attr_filter(self.entity.references, *args, **kwargs)

        if clusto_drivers:
            cdl = [clusto.get_driver_name(n) for n in clusto_drivers]
            result = (attr for attr in result if attr.entity.driver in cdl)

        if clusto_types:
            ctl = [clusto.get_type_name(n) for n in clusto_types]
            result = (attr for attr in result if attr.entity.type in ctl)


        return list(result)

    def referencers(self, *args, **kwargs):
        """Return the Things that reference _this_ Thing.

        Accepts the same arguments as references() but adds an instanceOf filter
        argument.
        """

        refs = [Driver(a.entity) for a in sorted(self.references(*args, **kwargs),
                                                 lambda x,y: cmp(x.attr_id,
                                                                 y.attr_id))]

        return refs


    def attr_keys(self, *args, **kwargs):

        return [x.key for x in self.attrs(*args, **kwargs)]

    def attr_key_tuples(self, *args, **kwargs):

        return [x.keytuple for x in self.attrs(*args, **kwargs)]

    def attr_items(self, *args, **kwargs):
        return self._itemize_attrs(self.attrs(*args, **kwargs))

    def add_attr(self, key, value=(), number=(), subkey=()):
        """add a key/value to the list of attributes

        if number is True, create an attribute with the next available
        otherwise number just gets passed to the Attribute constructor so it
        can be an integer or an sqlalchemy expression

        An optional subkey can also be specified. Subkeys don't affect
        numbering by default.
        """

        if isinstance(key, Attribute):
            raise Exception("Unsupported Operation.  You can no longer add an attribute directly")

        self._check_attr_name(key)
        if subkey:
            self._check_attr_name(subkey)

        if isinstance(value, Driver):
            value = value.entity

        if (number is ()) or (number is False):
            number = None
        if subkey is ():
            subkey = None


        self.expire(key=key)
        return self.entity.add_attr(key, value, subkey=subkey, number=number)


    def del_attrs(self, *args, **kwargs):
        "delete attribute with the given key and value optionally value also"

        clusto.flush()
        try:
            clusto.begin_transaction()
            for i in self.attr_query(*args, **kwargs):
                i.delete()
            clusto.commit()
            self.expire(*args, **kwargs)
        except Exception, x:
            clusto.rollback_transaction()
            raise x


    def set_attr(self, key, value, number=False, subkey=None):
        """replaces all attributes with the given key"""
        self._check_attr_name(key)

        attrs = self.attrs(key=key, number=number, subkey=subkey)

        if len(attrs) > 1:
            raise DriverException("cannot set an attribute when args match more than one value")

        else:
            if len(attrs) == 1:
                if attrs[0].value == value:
                    return attrs[0]
                self.del_attrs(key=key, number=number, subkey=subkey)

            attr = self.add_attr(key, value, number=number, subkey=subkey)

        return attr


    def expire(self, *args, **kwargs):
        """Expires the memcache value (if using memcache) of this object"""

        key = None
        if 'key' in kwargs:
            key = kwargs['key']
        subkey = None
        if 'subkey' in kwargs:
            subkey = kwargs['subkey']
        if clusto.SESSION.memcache:
            attrs = self.attrs(key=key, subkey=subkey)
            memcache_keys = []
            for attr in attrs:
                mk = self.name
                mk += '.%s' % attr.key
                if subkey is None:
                    memcache_keys.append(str(mk))
                if attr.subkey:
                    mk += '.%s' % attr.subkey
                memcache_keys.append(str(mk)) 
            memcache_keys = set(memcache_keys)
            for mk in memcache_keys:
                logging.debug('Expiring %s' % mk)
                clusto.SESSION.memcache.delete(mk)
        else:
            logging.info('Not using memcache, not expiring anything.')

    def has_attr(self, *args, **kwargs):
        """return True if this list has an attribute with the given key"""

        if self.attr_query(*args, **kwargs):
            return True

        return False

    def insert(self, thing):
        """Insert the given Enity or Driver into this Entity.  Such that:

        >>> A.insert(B)
        >>> (B in A)
        True


        """

        d = self.ensure_driver(thing,
                              "Can only insert an Entity or a Driver. "
                              "Tried to insert %s." % str(type(thing)))


        parent = thing.parents()

        if parent:
            raise TypeError("%s is already in %s and cannot be inserted into %s."
                            % (d.name, parent[0].entity.name, self.name))

        self.add_attr("_contains", d, number=True)

    def remove(self, thing):
        """Remove the given Entity or Driver from this Entity. Such that:

        >>> A.insert(B)
        >>> B in A
        True
        >>> A.remove(B)
        >>> B in A
        False

        """
        if isinstance(thing, Entity):
            d = Driver(Entity)
        elif isinstance(thing, Driver):
            d = thing
        else:
            raise TypeError("Can only remove an Entity or a Driver. "
                            "Tried to remove %s." % str(type(thing)))


        self.del_attrs("_contains", d, ignore_hidden=False)

    def content_attrs(self, *args, **kwargs):
        """Return the attributes referring to this Thing's contents

        """

        attrs = self.attrs("_contains", *args, **kwargs)

        return attrs

    def contents(self, *args, **kwargs):
        """Return the contents of this Entity.  Such that:

        >>> A.insert(B)
        >>> A.insert(C)
        >>> A.contents()
        [B, C]

        """

        if 'search_children' in kwargs:
            search_children = kwargs.pop('search_children')
        else:
            search_children = False

        contents = [attr.value for attr in self.content_attrs(*args, **kwargs)]
        if search_children:
            for child in (attr.value for attr in self.content_attrs()):
                kwargs['search_children'] = search_children
                contents.extend(child.contents(*args, **kwargs))

        return contents

    def parents(self, **kwargs):
        """Return a list of Things that contain _this_ Thing. """

        search_parents = kwargs.pop('search_parents', False)

        if search_parents:
            parents=self.parents(**kwargs)
            allparents = self.parents()
            for thing in allparents:
                allparents.extend(thing.parents())

            for thing in allparents:
                parents.extend(thing.parents(**kwargs))
        else:
            parents = self.referencers('_contains', **kwargs)

        return parents

    def siblings(self, parent_filter=None, parent_kwargs=None,
                 additional_pools=None, **kwargs):
        """Return a list of Things that have the same parents as me.

        parameters:
           parent_filter - a function used to filter out unwanted parents
           parent_kwargs - arguments to be passed to self.parents()
           additional_pools - a list of additional pools to use as sibling parents
           **kwargs - arguments to clusto.get_from_pools()
        """

        if parent_kwargs is None:
            parent_kwargs = dict(clusto_types=[clusto.drivers.Pool])

        parents = self.parents(**parent_kwargs)

        if parent_filter:
            parents = filter(parent_filter, parents)

        if additional_pools:
            parents.extend(additional_pools)

        return [s for s in clusto.get_from_pools(parents, search_children=False, **kwargs) if s != self]

    @classmethod
    def get_by_attr(cls, *args, **kwargs):
        """Get list of Drivers that have by attributes search """

        attrlist = cls.do_attr_query(*args, **kwargs)

        objs = [Driver(x.entity) for x in attrlist]

        return objs



    @property
    def name(self):
        return self.entity.name


    def update_metadata(self, *args, **kwargs):
        """Update entity's metadata"""
        pass



class ClustoException(Exception):
    """base clusto exception"""
    pass

class DriverException(ClustoException):
    """exception for driver errors"""
    pass

class ConnectionException(ClustoException):
    """exception for operations related to connecting two Things together"""
    pass


class NameException(ClustoException):
    """exception for invalid entity or attribute names"""
    pass


class ResourceException(ClustoException):
    """exception related to resources"""
    pass

class ResourceNotAvailableException(ResourceException):
    pass

class ResourceTypeException(ResourceException):
    pass


class PoolException(ClustoException):
    pass

class TransactionException(ClustoException):
    pass



from clusto.drivers.base import Driver
from clusto.schema import *

from clusto.exceptions import PoolException

from itertools import imap, chain

class Pool(Driver):
    """
    A Pool is used to group Entities into a collection that shares attributes.

    Pools
    """

    _driver_name = "pool"
    _clusto_type = "pool"


    def insert(self, thing):
        """Insert the given Enity or Driver into this Entity.

        Such that:

        >>> A.insert(B)
        >>> (B in A)
        True

        A given entity can only be in a Pool one time.
        """

        d = self.ensure_driver(thing,
                               "Can only insert an Entity or a Driver. "
                               "Tried to insert %s." % str(type(thing)))

        if d in self:
            raise PoolException("%s is already in pool %s." % (d, self))

        self.add_attr("_contains", d, number=True)


    def is_parent(self, thing):
        """
        Is this pool the parent of the given entity
        """

        d = self.ensure_driver(thing,
                               "Can only be the parent of a Driver or Entity.")

        return self in d.contents()

    @classmethod
    def get_pools(cls, obj, allPools=True):

        d = cls.ensure_driver(obj, "obj must be either an Entity or a Driver.")


        pools = [Driver(a.entity) for a in d.parents()
                 if isinstance(Driver(a.entity), Pool)]

        if allPools:
            for i in pools:
                pools.extend(Pool.get_pools(i, allPools=True))

        return pools

class ExclusivePool(Pool):
    _driver_name = "exclusive_pool"

    def insert(self, thing):
        """Insert the given Enity or Driver into this Entity.

        Such that:

        >>> A.insert(B)
        >>> (B in A)
        True

        A given entity can only be inserted into an ExclusivePool if
        it is in NO other pools.
        """

        pools = Pool.get_pools(thing)
        if pools:
            raise PoolException("%s is already in pools %s, cannot insert "
                                "exclusively." % (thing, pools))

        Pool.insert(self, thing)

class UniquePool(Pool):
    _driver_name = "unique_pool"

    def insert(self, thing):
        """Insert the given Enity or Driver into this Entity.

        Such that:

        >>> A.insert(B)
        >>> (B in A)
        True

        A given entity can only be in ONE UniquePool.
        """

        pools = thing.parents(clusto_drivers=[self._driver_name])
        if pools:
            raise PoolException("%s is already in UniquePool(s) %s." %
                                (thing, pools))

        Pool.insert(self, thing)




import clusto
from clusto.test import testbase

from clusto.drivers import Driver, PortMixin
from clusto.exceptions import ConnectionException

class TestDev1(Driver, PortMixin):
    _clusto_type = "test1"
    _driver_name = "test1driver"

    _portmeta = {'a' : {'numports':5},
                 'b' : {'numports':1}}

class TestDev2(Driver, PortMixin):

    _clusto_type = "test2"
    _driver_name = "test2driver"

    _portmeta = {'a' : {'numports':4},
                 'z' : {'numports':1}}

class PortLess(Driver):
    _clusto_type = "portless"
    _driver_name = "portless"

class PortTests(testbase.ClustoTestBase):
    """Test the port framework"""

    
    def data(self):
        
        a = TestDev1('t1')
        b = TestDev2('t2')
        c = PortLess('p')

    def testPortTypes(self):
        
        t1, t2, p = map(clusto.get_by_name, ['t1', 't2', 'p'])

        
        self.assertEqual(sorted(t1.port_types), 
                         sorted(['a', 'b']))

        self.assertEqual(sorted(t2.port_types), 
                         sorted(['a', 'z']))

    def testPortExists(self):
        
        t1, t2, p = map(clusto.get_by_name, ['t1', 't2', 'p'])

        self.assertTrue(t1.port_exists('a', 3))
        self.assertTrue(t1.port_exists('a', 1))
        self.assertFalse(t1.port_exists('a', 6))
        self.assertFalse(t1.port_exists('z', 4))

        self.assertTrue(t2.port_exists('z', 1))
        self.assertFalse(t2.port_exists('z', 2))

    def testPortsConnectable(self):
        
        t1, t2, p = map(clusto.get_by_name, ['t1', 't2', 'p'])

        self.assertTrue(t1.ports_connectable('a', 1, t2, 3))
        self.assertFalse(t1.ports_connectable('a', 1, t2, 5))
        self.assertFalse(t1.ports_connectable('b', 1, t2, 1))

    def testPortFree(self):
        
        t1, t2, p = map(clusto.get_by_name, ['t1', 't2', 'p'])

        self.assertTrue(t1.port_free('a', 2))

        t1.connect_ports('a', 1, t2, 1)

        self.assertFalse(t1.port_free('a', 1))

    def testConnectPorts(self):
        
        t1, t2, p = map(clusto.get_by_name, ['t1', 't2', 'p'])

        t1.connect_ports('a', 1, t2, 3)

        
        self.assertEqual(t2, t1.get_connected('a', 1))
        self.assertEqual(t1, t2.get_connected('a', 3))

        self.assertEqual(None, t1.get_connected('b', 1))


        # try to work with ports that don't exist
        self.assertRaises(ConnectionException, t2.get_connected, 'b', 3)
        self.assertRaises(ConnectionException, t2.connect_ports, 'b', 2, t1, 7)
        self.assertRaises(ConnectionException, t2.connect_ports, 'z', 1, t1, 1)


        # try to connect ports that are already connected but in the reverse order
        self.assertRaises(ConnectionException, t2.connect_ports, 'a', 4, t1, 1)

        # try to connect to a device that doesn't have ports
        self.assertRaises(ConnectionException, t1.connect_ports, 'a', 2, p, 1)

        # try to connect non integer ports
        self.assertRaises(TypeError, t1.connect_ports, 'a', '2', p, 1)

        # try to connect negative ports
        self.assertRaises(TypeError, t1.connect_ports, 'a', -2, p, 1)

                          

    def testDisconnectPort(self):

        t1, t2, p = map(clusto.get_by_name, ['t1', 't2', 'p'])

        t1.connect_ports('a', 1, t2, 3)

        self.assertEqual(t2, t1.get_connected('a', 1))
        
        t2.disconnect_port('a', 3)

        self.assertEqual(None, t1.get_connected('a', 1))
        self.assertEqual(None, t2.get_connected('a', 3))

        self.assertTrue(t1.port_free('a', 1))
        self.assertTrue(t2.port_free('a', 3))

        
    def testPortAttrs(self):

        t1, t2, p = map(clusto.get_by_name, ['t1', 't2', 'p'])

        t1.set_port_attr('a', 1, 'macaddr', 'foo')
        self.assertEqual('foo', t1.get_port_attr('a', 1, 'macaddr'))

        self.assertRaises(ConnectionException, 
                          t2.set_port_attr, 'j', 3, 'foo', 'bar')

        self.assertEqual(None, t2.get_port_attr('z', 1, 'mac'))
        t2.set_port_attr('z', 1, 'mac', 'bar')
        self.assertEqual('bar', t2.get_port_attr('z', 1, 'mac'))
        t2.del_port_attr('z', 1, 'mac')
        self.assertEqual(None, t2.get_port_attr('z', 1, 'mac'))
        


    def testPortInfo(self):
        
        t1, t2, p = map(clusto.get_by_name, ['t1', 't2', 'p'])

        self.assertEqual(sorted([('a', 1, None, None,),
                                 ('a', 2, None, None,),
                                 ('a', 3, None, None,),
                                 ('a', 4, None, None,),
                                 ('a', 5, None, None,),
                                 ('b', 1, None, None,),]),
                         sorted(t1.port_info_tuples))

        
        t1.connect_ports('a', 2, t2, 1)

        self.assertEqual(sorted([('a', 2, None, None,),
                                 ('a', 1, t1, 2,),
                                 ('a', 3, None, None,),
                                 ('a', 4, None, None,),
                                 ('z', 1, None, None,),]),
                         sorted(t2.port_info_tuples))

        self.assertEqual(t1, t2.port_info['a'][1]['connection'])
        self.assertEqual(2, t2.port_info['a'][1]['otherportnum'])

        self.assertEqual(None, t2.port_info['a'][3]['connection'])
        self.assertEqual(None, t2.port_info['z'][1]['otherportnum'])

        self.assertEqual(sorted([('a', 2),
                                 ('a', 3),
                                 ('a', 4),
                                 ('z', 1),]),
                         sorted(t2.free_ports))

        self.assertEqual(sorted(['a', 'b']),
                         sorted(t1.port_types))

        self.assertEqual(sorted(['a', 'z']),
                         sorted(t2.port_types))


    def testConnectedPorts(self):

        t1, t2, p = map(clusto.get_by_name, ['t1', 't2', 'p'])

        for i in [t1, t2]:
            for t in i.port_types:
                self.assertEqual([], i.connected_ports[t])

        t1.connect_ports('a', 1, t2, 3)
        t2.connect_ports('a', 2, t1, 2)
        
        self.assertEqual([1, 2], t1.connected_ports['a'])
        



#!/usr/bin/env python
# -*- mode: python; sh-basic-offset: 4; indent-tabs-mode: nil; coding: utf-8 -*-
# vim: tabstop=4 softtabstop=4 expandtab shiftwidth=4 fileencoding=utf-8

from IPy import IP
import argparse
import sys

import clusto
from clusto import drivers
from clusto import script_helper


class ListPool(script_helper.Script):
    '''
    Lists servers that are part of two given pools.
    '''

    def __init__(self):
        script_helper.Script.__init__(self)

    def _add_arguments(self, parser):
        parser.add_argument('--names', default=False, action='store_true',
            help='Print names instead of ip addresses (defaults to False)')
        parser.add_argument('--recursive', default=False, action='store_true',
            help='Search resursively on both pools (defaults to False)')
        parser.add_argument('--type', default=None,
            help='Restrict results to the given type')
        parser.add_argument('pool', nargs='+', metavar='pool',
            help='Pool(s) to query')

    def run(self, args):
        serverset = None
        for pool in args.pool:
            pool = clusto.get_by_name(pool)
            contents = pool.contents(search_children=args.recursive)
            if args.type:
                contents = [x for x in contents if x.type == args.type]
            contents = set(contents)

            if serverset is None:
                serverset = contents
            else:
                serverset = serverset.intersection(pool)

        for server in serverset:
            if args.names:
                print server.name
            else:
                try:
                    ip = server.get_ips()
                    ip.sort(key=lambda x: IP(x))
                except Exception, e:
                    self.debug(e)
                    ip = None

                if ip:
                    print ' '.join(ip)
                else:
                    print server.name


def main():
    lp, args = script_helper.init_arguments(ListPool)
    return(lp.run(args))

if __name__ == '__main__':
    sys.exit(main())




#!/usr/bin/env python
try:
    import simplejson as json
except ImportError:
    import json

from webob import Request, Response
from traceback import format_exc
from urllib import unquote_plus
import new
import re
import os

from clusto.drivers import Driver, IPManager
import clusto.script_helper
import clusto

from clusto.services.config import conf, get_logger
log = get_logger('clusto.http', 'INFO')

conf = clusto.script_helper.load_config(os.environ.get('CLUSTOCONFIG', '/etc/clusto/clusto.conf'))
clusto.connect(conf)


def unclusto(obj):
    '''
    Convert an object to a representation that can be safely serialized into
    JSON.
    '''
    if type(obj) in (str, unicode, int) or obj == None:
        return obj
    if isinstance(obj, clusto.Attribute):
        return {
            'key': obj.key,
            'value': unclusto(obj.value),
            'subkey': obj.subkey,
            'number': obj.number,
            'datatype': obj.datatype
        }
    if issubclass(obj.__class__, Driver):
        return '/%s/%s' % (obj.type, obj.name)
    return str(obj)


def dumps(request, obj, **kwargs):
    result = json.dumps(obj, indent=2, sort_keys=True, **kwargs)
    if 'callback' in request.params:
        callback = request.params['callback']
        result = '%s(%s)' % (callback, result)
        content_type = 'text/javascript'
    else:
        content_type = 'application/json'
    return Response(result, content_type=content_type, **kwargs)


def loads(request, obj):
    return json.loads(obj)


class EntityAPI(object):
    def __init__(self, obj):
        self.obj = obj
        self.url = '/%s/%s' % (self.obj.type, self.obj.name)

    def _dict(self):
        result = {
            'object': self.url,
            'driver': self.obj.driver,
        }

        attrs = []
        for x in self.obj.attrs():
            attrs.append(unclusto(x))
        result['attrs'] = attrs
        result['contents'] = [unclusto(x) for x in self.obj.contents()]
        result['parents'] = [unclusto(x) for x in self.obj.parents()]
        result['actions'] = [x for x in dir(self) if not x.startswith('_') and callable(getattr(self, x))]
        return result

    def addattr(self, request):
        '''
        Add an attribute to this object.

        Requires HTTP parameters "key" and "value"
        Optional parameters are "subkey" and "number"
        '''
        kwargs = dict(request.params.items())
        self.obj.add_attr(**kwargs)
        return self.show(request)

    def setattr(self, request):
        '''
        Set an attribute on this object.

        Requires HTTP parameters "key" and "value"
        Optional parameters are "subkey" and "number"
        '''
        kwargs = dict(request.params.items())
        if 'number' in kwargs:
            kwargs['number'] = int(kwargs['number'])
        self.obj.set_attr(**kwargs)
        return self.show(request)

    def delattr(self, request):
        '''
        Delete an attribute from this object.

        Requires HTTP parameter "key"
        '''
        self.obj.del_attrs(request.params['key'])
        return self.show(request)

    def attrs(self, request):
        '''
        Query attributes from this object.
        '''
        result = {
            'attrs': []
        }
        kwargs = dict(request.params.items())
        for attr in self.obj.attrs(**kwargs):
            result['attrs'].append(unclusto(attr))
        return dumps(request, result)

    def insert(self, request):
        '''
        Insert an object into this object

        Requires a "device" attribute, which is the absolute URL path to the
        object to be inserted.

        For example: /pool/examplepool/insert?object=/server/exampleserver
        '''
        device = request.params['object'].strip('/').split('/')[1]
        device = clusto.get_by_name(device)
        self.obj.insert(device)
        return self.show(request)

    def remove(self, request):
        '''
        Remove a device from this object.

        Requires a "device" attribute, which is the absolute URL path to the
        object to be removed.

        For example: /pool/examplepool/remove?object=/server/exampleserver
        '''
        device = request.params['object'].strip('/').split('/')[1]
        device = clusto.get_by_name(device)
        self.obj.remove(device)
        return self.show(request)

    def show(self, request):
        '''
        Returns attributes and actions available for this object.
        '''
        result = {}
        result['object'] = self.url
        result['driver'] = self.obj.driver

        attrs = []
        for x in self.obj.attrs():
            attrs.append(unclusto(x))
        result['attrs'] = attrs
        result['contents'] = [unclusto(x) for x in self.obj.contents()]
        result['parents'] = [unclusto(x) for x in self.obj.parents()]
        result['actions'] = [x for x in dir(self) if not x.startswith('_') and callable(getattr(self, x))]

        return dumps(request, result)


class PortInfoAPI(EntityAPI):
    def ports(self, request):
        '''
        Returns a list of ports within this object and any information about
        connections between those ports and other devices.
        '''
        result = {}
        result['object'] = self.url
        result['ports'] = []
        for port in self.obj.port_info_tuples:
            porttype, portnum, otherobj, othernum = [unclusto(x) for x in port]
            result['ports'].append({
                'type': porttype,
                'num': portnum,
                'other': otherobj,
                'othernum': othernum,
            })

        return dumps(request, result)

    def set_port_attr(self, request):
        kwargs = {}
        params = request.params
        for arg in ('porttype', 'portnum', 'key', 'value'):
            if not arg in params:
                return Response(status=400, body='Required argument "%s" is missing\n' % arg)
            kwargs[arg] = params[arg]

        try:
            kwargs['portnum'] = int(kwargs['portnum'])
        except ValueError:
            return Response(status=400, body='portnum must be an integer\n')

        self.obj.set_port_attr(**kwargs)
        return Response(status=200)

    def get_port_attr(self, request):
        kwargs = {}
        params = request.params
        for arg in ('porttype', 'portnum', 'key'):
            if not arg in params:
                return Response(status=400, body='Required argument "%s" is missing\n' % arg)
            kwargs[str(arg)] = params[arg]

        try:
            kwargs['portnum'] = int(kwargs['portnum'])
        except ValueError:
            return Response(status=400, body='portnum must be an integer\n')

        return dumps(request, self.obj.get_port_attr(**kwargs))


class RackAPI(EntityAPI):
    def insert(self, request):
        '''
        Insert a device into a rack.

        This method is overridden to require a "ru" parameter in addition to
        "device"

        Example: /rack/examplerack/insert?object=/server/exampleserver&ru=6
        '''
        device = request.params['object'].strip('/').split('/')[1]
        device = clusto.get_by_name(device)
        self.obj.insert(device, int(request.params['ru']))
        return self.contents(request)


class ResourceAPI(EntityAPI):
    def allocate(self, request):
        '''
        Allocate a new object of the given type
        '''
        driver = clusto.DRIVERLIST[request.params['driver']]
        device = self.obj.allocate(driver)
        return dumps(request, unclusto(device), status=201)


class QueryAPI(object):
    @classmethod
    def get_entities(self, request):
        kwargs = {'attrs': []}
        for k, v in request.params.items():
            if k in ('format', 'callback'): continue
            v = loads(request, unquote_plus(v))
            kwargs[str(k)] = v

        attrs = []
        for attr in kwargs['attrs']:
            attrs.append(dict([(str(k), v) for k, v in attr.items()]))
        kwargs['attrs'] = attrs

        result = [unclusto(x) for x in clusto.get_entities(**kwargs)]
        return dumps(request, result)

    @classmethod
    def get_by_name(self, request):
        if not 'name' in request.params:
            return Response(status=400, body='400 Bad Request\nYou must specify a "name" parameter\n')
        name = request.params['name']
        obj = clusto.get_by_name(name)
        api = EntityAPI(obj)
        return api.show(request)

    @classmethod
    def get(self, request):
        if not 'name' in request.params:
            return Response(status=400, body='400 Bad Request\nYou must specifiy a "name" parameter\n')
        name = request.params['name']

        result = clusto.get(name)
        if result is None:
            return Response(status=404, body='404 Not Found\n')

        result = [EntityAPI(x)._dict() for x in result]
        return dumps(request, result)

    @classmethod
    def get_from_pools(self, request):
        if not 'pools' in request.params:
            return Response(status=400, body='400 Bad Request\nYou must specify at least one pool\n')
        pools = request.params['pools'].split(',')

        if 'types' in request.params:
            clusto_types = request.params['types'].split(',')
        else:
            clusto_types = None

        result = [unclusto(x) for x in clusto.get_from_pools(pools, clusto_types)]
        return dumps(request, result)

    @classmethod
    def get_ip_manager(self, request):
        if not 'ip' in request.params:
            return Response(status=400, body='400 Bad Request\nYou must specify an IP\n')

        try:
            ipman = IPManager.get_ip_manager(request.params['ip'])
        except:
            return Response(status=404, body='404 Not Found\n')
        return dumps(request, unclusto(ipman))


class ClustoApp(object):
    def __init__(self):
        self.urls = [
            ('^/favicon.ico$',
                self.notfound),
            ('^/search$',
                self.search),
            ('^/query/(?P<querytype>[a-z_]+)',
                self.query_delegate),
            ('^/(?P<objtype>\w+)/(?P<name>[-\w0-9]+)/(?P<action>\w+)',
                self.action_delegate),
            ('^/(?P<objtype>\w+)/(?P<name>[-\w0-9]+)',
                self.action_delegate),
            ('^/(?P<objtype>\w+)',
                self.types_delegate),
            ('^/',
                self.default_delegate),
        ]
        self.urls = [(re.compile(pattern), obj) for pattern, obj in self.urls]

        self.types = {
            'server': PortInfoAPI,
            'consoleserver': PortInfoAPI,
            'networkswitch': PortInfoAPI,
            'powerstrip': PortInfoAPI,
            'rack': RackAPI,
            'resource': ResourceAPI,
        }

    def default_delegate(self, request, match):
        types = ['/' + x for x in clusto.typelist.keys()]
        types.sort()
        return dumps(request, types)

    def types_delegate(self, request, match):
        objtype = match.groupdict()['objtype']
        result = []
        for obj in clusto.get_entities(clusto_types=(objtype,)):
            result.append(unclusto(obj))
        return dumps(request, result)

    def action_delegate(self, request, match):
        if request.method == 'GET':
            return self.get_action(request, match)

        if request.method == 'POST':
            return self.post_action(request, match)

        if request.method == 'DELETE':
            return self.delete_action(request, match)

        return Response(status=501, body='501 Not Implemented\n')

    def query_delegate(self, request, match):
        querytype = match.groupdict()['querytype']

        if hasattr(QueryAPI, querytype):
            method = getattr(QueryAPI, querytype)
            return method(request)
        else:
            return Response(status=400, body='400 Bad Request\nThere is no such query\n')

    def post_action(self, request, match):
        name = request.path_info.strip('/')
        if name.count('/') != 1:
            return Response(status=400, body='400 Bad Request\nYou may only create objects, not types or actions\n')
        objtype, objname = name.split('/', 1)

        try:
            obj = clusto.get_by_name(objname)
            if obj:
                return Response(status=409, body='409 Conflict\nObject already exists\n')
        except LookupError: pass

        obj = clusto.driverlist[objtype](objname)

        obj = EntityAPI(obj)
        response = obj.show(request)
        response.status = 201
        return response

    def delete_action(self, request, match):
        name = request.path_info.strip('/')
        if name.count('/') != 1:
            return Response(status=400, body='400 Bad Request\nYou may only delete objects, not types or actions\n')
        objtype, objname = name.split('/', 1)

        try:
            obj = clusto.get_by_name(objname)
        except LookupError:
            return Response(status=404, body='404 Not Found\n')

        clusto.delete_entity(obj.entity)
        return Response(status=200, body='200 OK\nObject deleted\n')

    def get_action(self, request, match):
        group = match.groupdict()
        try:
            obj = clusto.get_by_name(group['name'])
        except LookupError:
            return Response(status=404, body='404 Not Found\n')

        if obj.type != group['objtype']:
            response = Response(status=302)
            response.headers.add('Location', str('/%s/%s/%s' % (obj.type, obj.name, group.get('action', ''))))
            return response

        action = group.get('action', 'show')
        handler = self.types.get(group['objtype'], EntityAPI)
        if not obj:
            obj = clusto.get_by_name(group['name'])
        h = handler(obj)
        if hasattr(h, action):
            f = getattr(h, action)
            response = f(request)
        else:
            response = Response(status=404, body='404 Not Found\nInvalid action\n')
        return response

    def search(self, request, match):
        query = request.params.get('q', None)
        if not query:
            return Response(status=400, body='400 Bad Request\nNo query specified\n')

        result = []
        for obj in clusto.get_entities():
            if obj.name.find(query) != -1:
                result.append(unclusto(obj))
        return dumps(request, result)

    def notfound(self, request, match):
        return Response(status=404)

    def __call__(self, environ, start_response):
        request = Request(environ)
        response = Response(status=404, body='404 Not Found\nUnmatched URL\n')

        for pattern, handler in self.urls:
            match = pattern.match(request.path_info)
            if match:
                try:
                    response = handler(request, match)
                except:
                    response = Response(status=500, body=format_exc(),
                                        content_type='text/plain')
                break
        return response(environ, start_response)


application = ClustoApp()


if __name__ == '__main__':
    from wsgiref.simple_server import make_server
    server = make_server('0.0.0.0', 9996, application)
    log.info('clusto httpd loaded using wsgiref (debug mode)')
    server.serve_forever()



"""
IPMixin is a basic mixin to be used by devices that can be assigned IPs
"""

import re

import clusto

from clusto.drivers.resourcemanagers import IPManager

from clusto.exceptions import ConnectionException,  ResourceException


class IPMixin:

    def add_ip(self, ip=None, ipman=None):

        if not ip and not ipman:
            raise ResourceException('If no ip is specified then an ipmanager must be specified')

        elif ip:
            
            if not ipman:
                ipman = IPManager.get_ip_manager(ip)

            return ipman.allocate(self, ip)
        else:
            return ipman.allocate(self)
            
            
        
    def has_ip(self, ip):

        ipman = IPManager.get_ip_manager(ip)

        return self in ipman.owners(ip)

    def get_ips(self):
        """Get a list of IPs for this Entity in ipstring format"""

        return self.attr_values(IPManager._attr_name, subkey='ipstring')
        

    def bind_ip_to_osport(self, ip, osportname, ipman=None, porttype=None, portnum=None):
        """bind an IP to an os port and optionally also asign the os port name
        to a physical port

        If the given ip is already allocated to this device then use it.  If
        it isn't, try to allocate it from a matching IPManager.

        
        """

        if (porttype != None) ^ (portnum != None):
                raise Exception("both portype and portnum need to be specified or set to None")
            
        try:
            clusto.begin_transaction()

            if not self.has_ip(ip):
                if not ipman:
                    ipman = IPManager.get_ip_manager(ip)

                ipman.allocate(self, ip)

                clusto.flush()
            else:
                ipman = IPManager.get_ip_manager(ip)

            number = ipman.get_resource_number(self, ip)
            ipattrs = ipman.get_resource_attrs(self, ip, number=number)

            if porttype is not None and portnum is not None:
                self.set_port_attr(porttype, portnum, 'osportname', osportname)

            self.set_attr(ipattrs[0].key,
                         number=ipattrs[0].number,
                         subkey='osportname',
                         value=osportname)

            clusto.commit()
        except Exception, x:
            clusto.rollback_transaction()
            raise x
        


        



            
        



import unittest
from clusto.test import testbase

import clusto
from clusto.schema import *
from clusto.drivers.base import *
from clusto.drivers import BasicDatacenter
from sqlalchemy.exc import InvalidRequestError

class TestClustoVersioning(testbase.ClustoTestBase):

    def testGetFirstVersionNumber(self):

        curver = clusto.get_latest_version_number()
        self.assertEqual(curver, 2)

    def testVersionIncrementing(self):

        curver = clusto.get_latest_version_number()

        e1 = Entity('e1')
        e2 = Entity('e2')

        self.assertEqual(clusto.get_latest_version_number(), curver + 2)

    def testVersionIncrementWithAttrs(self):

        curver = clusto.get_latest_version_number()
        
        e1 = Entity('e1')
        e2 = Entity('e2')

        e1.add_attr('foo', 2)

        
        self.assertEqual(clusto.get_latest_version_number(), curver + 3)
        
        
    def testDeleteVersion(self):

        curver = clusto.get_latest_version_number()

        e1 = Entity('e1')
        etest = clusto.get_by_name('e1')
        e1.delete()


        self.assertRaises(LookupError, clusto.get_by_name, 'e1')

        e1a = Entity('e1')

        etest = clusto.get_by_name('e1')

        self.assertEqual(etest.entity.version, curver+3)


    def testViewOldVersion(self):

        curver = clusto.get_latest_version_number()

        e1 = Entity('e1')
        e2 = Entity('e2')
        e3 = Entity('e3')

        self.assertEqual(Entity.query().filter(Entity.name.like(u'e%')).count(),
                         3)

        SESSION.clusto_version = curver

        self.assertEqual(Entity.query().filter(Entity.name.like(u'e%')).count(),
                         0)

        SESSION.clusto_version = clusto.working_version()

        self.assertEqual(Entity.query().filter(Entity.name.like(u'e%')).count(),
                         3)

        SESSION.clusto_version = curver + 1

        self.assertEqual(Entity.query().filter(Entity.name.like(u'e%')).count(),
                         1)


        SESSION.clusto_version = curver + 2

        self.assertEqual(sorted([e1,e2]),
                         Entity.query().filter(Entity.name.like(u'e%')).all())

    def testOldVersionsOfAttributes(self):

        curver = clusto.get_latest_version_number()

        e1 = Entity('e1')
        e2 = Entity('e2')

        e1.add_attr('foo', 1)
        e1.add_attr('foo2', 2)
        e1.add_attr('foo3', 3)

        SESSION.clusto_version = curver + 3

        self.assertEqual(len(list(e1.attrs)), 1)
        
        e = Entity.query().filter_by(name=u'e1').one()

        SESSION.clusto_version = curver + 4

        self.assertEqual(sorted([a.key for a in e.attrs]),
                         sorted(['foo', 'foo2']))


    def testAttributesImmutable(self):

        e1 = Entity('e1')
        e1.add_attr('foo', 1)

        a = e1.attrs[0]

        self.assertRaises(Exception, setattr, a.value, 2)

        
    def testEntityImmutable(self):

        e1 = Entity('e1')

        self.assertRaises(Exception, setattr, e1.driver, 'foo')

    def testEntityRename(self):

        curver = clusto.get_latest_version_number()

        e1 = Entity('e1')

        e1.add_attr('foo',1)
        e1.add_attr('foo',2)

        e1attrs = [a.to_tuple for a in e1.attrs]
        
        midver = clusto.get_latest_version_number()

        clusto.rename('e1', 't1')

        postrenamever = clusto.get_latest_version_number()
        
        t1 = clusto.get_by_name('t1')

        self.assertEqual(sorted(e1attrs),
                         sorted([a.to_tuple for a in t1.entity.attrs]))

        
        t1.del_attrs('foo', 2)

        self.assertRaises(LookupError, clusto.get_by_name, 'e1')

        self.assertEqual(sorted(t1.attrs('foo',1)),
                         sorted(t1.attrs()))

        SESSION.clusto_version = midver

        self.assertRaises(LookupError, clusto.get_by_name, 't1')

        e = clusto.get_by_name('e1')

        self.assertEqual(sorted(e1attrs),
                         sorted([a.to_tuple for a in e.attrs()]))

        
        for a in e.attrs():
            self.assertEqual(e.entity.deleted_at_version,
                             a.deleted_at_version)

        SESSION.clusto_version = postrenamever

        self.assertEqual(e.entity.deleted_at_version,
                         t1.entity.version)

        for a in t1.attrs():
            self.assertEqual(e.entity.deleted_at_version,
                             a.version)

    def testPoolRename(self):

        
        curver = clusto.get_latest_version_number()

        e1 = Entity('e1')
        e2 = Entity('e2')

        p1 = clusto.drivers.Pool('p1')

        p1.insert(e1)
        p1.insert(e2)

        self.assertEqual(sorted([e1,e2]),
                         sorted((d.entity for d in p1.contents())))

        clusto.rename('p1', 'p1renamed')

        p1renamed = clusto.get_by_name('p1renamed')

        self.assertEqual(sorted([e1,e2]),
                         sorted((d.entity for d in p1renamed.contents())))



    def testSetAttrIncrementsVersion(self):

        curver = clusto.get_latest_version_number()
        
        d = clusto.drivers.Driver('d1')

        self.assertEqual(curver + 1, clusto.get_latest_version_number())

        Attribute(d.entity, 'cat', 'baz')

        clusto.begin_transaction()
        SESSION.clusto_description = "TEST"

        #Attribute(d.entity, 'foo', 'cat')
        #SESSION.add(Attribute(d.entity, 'coo', 'daa'))

        Attribute.query().all()

        a = Attribute(d.entity, 'foo', 'bar')

        clusto.commit()
        #clusto.rollback_transaction()
            
        #self.assertEqual(['bar', 'bar'], d.attr_values('foo'))

        self.assert_('bar' in d.attr_values('foo'))
        self.assertEqual(curver + 3, clusto.get_latest_version_number())
        
    def testEmptyCommits(self):


        server = clusto.drivers.BasicServer('s1')

        curver = clusto.get_latest_version_number()
        server.attrs()

        self.assertEqual(curver, clusto.get_latest_version_number())
        

        try:
            clusto.begin_transaction()

            SESSION.clusto_description = "TEST"

            Entity.query().all()

            clusto.commit()

        except:
            
            clusto.rollback_transaction()

        self.assertEqual(curver, clusto.get_latest_version_number())

        self.assertEqual([], server.attr_values('foo'))



import clusto
from clusto.schema import Attribute

from clusto.drivers.base import ResourceManager, ResourceTypeException, Driver
from clusto.exceptions import ResourceNotAvailableException, ResourceException

import IPy

class IPManager(ResourceManager):
    """Resource Manager for IP spaces
    
    roughly follows the functionality available in IPy
    """


    _driver_name="ipmanager"
    _clusto_type="ipmanager"

    _properties = {'gateway': None,
                   'netmask': '255.255.255.255',
                   'baseip': None }

    _attr_name = "ip"

    _int_ip_const = 2147483648


    @classmethod
    def _ipy_to_int(cls, ipy):
        return int(ipy.int() - cls._int_ip_const)

    @classmethod
    def _int_to_ipy(cls, num):
        return IPy.IP(num + cls._int_ip_const)
    
    @property
    def ipy(self):
        if not hasattr(self, '__ipy'):

            self.__ipy = IPy.IP(''.join([u'%s' % str(self.baseip), '/', 
                                u'%s' % self.netmask]), make_net=True)


        return self.__ipy

    def ensure_type(self, resource, number=True, thing=None):
        """check that the given ip falls within the range managed by this manager"""

        try:
            if isinstance(resource, int):
                ip = self._int_to_ipy(resource)
            else:
                ip = IPy.IP(resource)
        except ValueError:
            raise ResourceTypeException("%s is not a valid ip."
                                        % resource)

        if self.baseip and (ip not in self.ipy):
            raise ResourceTypeException(u"The ip %s is out of range for this IP manager.  Should be in %s/%s"
                                        % (str(ip), self.baseip, self.netmask))


        return (self._ipy_to_int(ip), number)


    def additional_attrs(self, thing, resource, number):

        resource, number = self.ensure_type(resource, number)

        thing.add_attr(self._attr_name, number=number, subkey='ipstring', value=str(self._int_to_ipy(resource)))
        
                     
    def allocator(self, thing=None):
        """allocate IPs from this manager"""

        if self.baseip is None:
            raise ResourceTypeException("Cannot generate an IP for an ipManager with no baseip")

        lastip = self.attr_query('_lastip')
                
        if not lastip:
            # I subtract self._int_ip_const to keep in int range
            startip=int(self.ipy.net().int() + 1) - self._int_ip_const 
        else:
            startip = lastip[0].value


        
        ## generate new ips the slow naive way
        nextip = int(startip)
        if self.gateway:
            gateway = IPy.IP(self.gateway).int() - self._int_ip_const
        else:
            gateway = None
        endip = self.ipy.broadcast().int() - self._int_ip_const

        for i in range(2):
            while nextip < endip:

                if nextip == gateway:
                    nextip += 1
                    continue

                if self.available(nextip):
                    self.set_attr('_lastip', nextip)
                    return self.ensure_type(nextip, True)
                else:
                    nextip += 1
            
            # check from the beginning again in case an earlier ip
            # got freed
                    
            nextip = int(self.ipy.net().int() + 1)
            
        raise ResourceNotAvailableException("out of available ips.")

    @classmethod
    def get_ip_manager(cls, ip):
        """return a valid ip manager for the given ip.

        @param ip: the ip
        @type ip: integer, string, or IPy object

        @return: the appropriate IP manager from the clusto database
        """

        ipman = None
        if isinstance(ip, Attribute):
            ipman = ip.entity
            return Driver(ipman)

        for ipmantest in clusto.get_entities(clusto_types=[cls]):
            try:
                ipmantest.ensure_type(ip)
            except ResourceTypeException:
                continue

            ipman = Driver(ipmantest)
            break
        

        if not ipman:
            raise ResourceException(u"No resource manager for %s exists."
                                    % str(ip))
        
        return ipman
        
    @classmethod
    def get_ips(cls, device):

        ret = [str(cls._int_to_ipy(x.value))
               for x in cls.resources(device)]

        return ret

    @classmethod
    def get_devices(self, ip):
        subnet = IPManager.get_ip_manager(ip)
        return subnet.owners(ip)



import unittest
from clusto.test import testbase

import clusto
from clusto.schema import *
from clusto.drivers.base import *
from clusto.drivers import BasicDatacenter, Pool, BasicServer, IPManager
from sqlalchemy.exc import InvalidRequestError
from clusto.exceptions import TransactionException


class TestClustoPlain(testbase.ClustoTestBase):

    def testInitClustoIdempotent(self):

        clusto.init_clusto()
        clusto.init_clusto()
        clusto.init_clusto()
        clusto.init_clusto()

        self.assertEqual(SESSION.query(ClustoVersioning).count(), 2)



class TestClusto(testbase.ClustoTestBase):
    def data(self):

        Entity('e1')
        Entity('e2')
        Entity('e3')

        clusto.flush()


    def testClustoMeta(self):

        cm = clusto.get_by_name('clustometa')

        self.assertEqual(cm.schemaversion, VERSION)

    def testGetByName(self):

        e1 = Entity.query().filter_by(name=u'e1').one()

        q = clusto.get_by_name('e1')

        self.assertEqual(q, e1)

        self.assertEqual(q.name, 'e1')

    def testGetByNames(self):

        names = ['e3','e1','e2']
        entities = [clusto.get_by_name(x) for x in names]

        self.assertEqual(entities, clusto.get_by_names(names))

        self.assertEqual([clusto.get_by_name('e3'),
                          None,
                          clusto.get_by_name('e1')],
                         clusto.get_by_names(['e3', 'shouldfail', 'e1']))


    def testSimpleRename(self):

        clusto.rename('e1', 'f1')

        q = Entity.query()

        self.assertEqual(q.filter_by(name=u'e1').count(), 0)

        self.assertEqual(q.filter_by(name=u'f1').count(), 1)


    def testChangeDriver(self):

        d = Driver('d1')
        d.add_attr('foo', 1)

        self.assertEqual(d.driver, Driver._driver_name)

        self.assertRaises(Exception, setattr, d.entity, driver, 'foo')

        clusto.change_driver(d.name, BasicServer)

        self.assertRaises(DriverException, clusto.change_driver, d.name, str)

        d = clusto.get_by_name('d1')

        self.assertEqual(d.driver, BasicServer._driver_name)
        self.assertTrue(isinstance(d, BasicServer))

        self.assertEqual(d.attr_value('foo'), 1)


    def testBadRollback(self):

        clusto.begin_transaction()

        d1 = Entity('d1')

        clusto.get_by_name('d1')

        d2 = Entity('d2')
        clusto.commit()

        self.assertRaises(TransactionException, clusto.rollback_transaction)

    def testTransactionRollback(self):

        clusto.begin_transaction()

        d1 = Entity('d1')

        clusto.get_by_name('d1')

        d2 = Entity('d2')
        clusto.rollback_transaction()


        self.assertRaises(LookupError, clusto.get_by_name, 'd1')

    def testTransactionRollback2(self):

        try:
            clusto.begin_transaction()

            c1 = Entity('c1')

            raise Exception()
        except Exception:

            clusto.rollback_transaction()

        c2 = Entity('c2')

        self.assertRaises(LookupError, clusto.get_by_name, 'c1')
        clusto.get_by_name('c2')

    def testTransactionRollback3(self):

        d1 = Entity('d1')

        clusto.begin_transaction()
        d2 = Entity('d2')
        clusto.rollback_transaction()

        clusto.get_by_name('d1')
        self.assertRaises(LookupError, clusto.get_by_name, 'd2')

    def testTransactionRollback4(self):

        d1 = Driver('d1')

        try:
            clusto.begin_transaction()

            d2 = Driver('d2')

            try:
                clusto.begin_transaction()
                d2.add_attr('foo', 'bar')

                clusto.commit()

            except:
                clusto.rollback_transaction()

            d1.add_attr('foo2', 'bar2')

            raise Exception()
            clusto.commit()
        except:
            clusto.rollback_transaction()

        self.assertEqual(d1.attrs(), [])
        self.assertRaises(LookupError, clusto.get_by_name, 'd2')


    def testTransactionCommit(self):

        try:
            clusto.begin_transaction()

            c1 = Entity('c1')
            clusto.commit()
        except Exception:
            clusto.rollback_transaction()

        clusto.get_by_name('c1')


    def testGetEntities(self):

        d1 = Driver('d1')
        dv1 = Device('dv1')
        Location('l1')
        BasicDatacenter('dc1')


        namelist = ['e1', 'e2', 'dv1']

        self.assertEqual(sorted([n.name
                                 for n in clusto.get_entities(names=namelist)]),
                         sorted(namelist))

        dl = [Driver]
        self.assertEqual(sorted([n.name
                                 for n in clusto.get_entities(clusto_drivers=dl)]),
                         sorted(['d1','e1','e2','e3']))


        tl = [Location, BasicDatacenter]
        self.assertEqual(sorted([n.name
                                 for n in clusto.get_entities(clusto_types=tl)]),
                         sorted(['l1','dc1']))

        p1 = Pool('p1')
        p2 = Pool('p2')
        p3 = Pool('p3')
        p4 = Pool('p4')

        s1 = BasicServer('s1')
        s2 = BasicServer('s2')
        s3 = BasicServer('s3')

        p1.insert(s1)
        p1.insert(s2)

        p2.insert(s2)
        p2.insert(s3)
        p2.insert(d1)

        p3.insert(s3)
        p3.insert(d1)
        p3.insert(s1)

        p4.insert(p3)


        self.assertEqual(sorted([s2,s3]),
                         sorted(clusto.get_from_pools(pools=[p2],
                                                      clusto_types=[BasicServer])))

        self.assertEqual(sorted([s2]),
                         sorted(clusto.get_from_pools(pools=[p2, 'p1'],
                                                      clusto_types=[BasicServer])))
        self.assertEqual(sorted([s3]),
                         sorted(clusto.get_from_pools(pools=['p2', 'p3'],
                                                      clusto_types=[BasicServer])))

        self.assertEqual(sorted([s1]),
                         sorted(clusto.get_from_pools(pools=['p4', 'p1'],
                                                      clusto_types=[BasicServer])))

    def testGetEntitesWithAttrs(self):

        d1 = Driver('d1')
        d2 = Driver('d2')
        d3 = Driver('d3')
        d4 = Driver('d4')

        d1.add_attr('k1', 'test')
        d2.add_attr('k1', 'testA')

        d1.add_attr('k2', number=1, subkey='A', value=67)
        d3.add_attr('k3', number=True, value=d4)



        self.assertEqual(clusto.get_entities(attrs=[{'key':'k2'}]),
                         [d1])


        self.assertEqual(sorted(clusto.get_entities(attrs=[{'key':'k1'}])),
                         sorted([d1,d2]))


        self.assertEqual(sorted(clusto.get_entities(attrs=[{'value':d4}])),
                         [d3])


        self.assertEqual(clusto.get_entities(attrs=[{'value':67}]),
                         [d1])

        self.assertEqual(sorted(clusto.get_entities(attrs=[{'number':0}])),
                         sorted([d3]))

        self.assertEqual(clusto.get_entities(attrs=[{'subkey':'A'},
                                                   {'value':'test'}]),
                         [d1])

    def testGet(self):
        s1 = BasicServer('s1')
        s2 = BasicServer('s2')
        s3 = BasicServer('s3')
        ipm = IPManager('testnet', baseip='10.0.0.0', netmask='255.255.255.0')

        s1.set_attr(key='system', subkey='serial', value='P0000000000')
        s2.set_port_attr('nic-eth', 1, 'mac', '00:11:22:33:44:55')
        s3.bind_ip_to_osport('10.0.0.1', 'eth0')

        self.assertEqual(clusto.get('s1')[0], s1)
        self.assertEqual(clusto.get('00:11:22:33:44:55')[0], s2)
        self.assertEqual(clusto.get('10.0.0.1')[0], s3)
        self.assertEqual(clusto.get('P0000000000')[0], s1)
        self.assertEqual(clusto.get('foo'), None)
        self.assertRaises(ValueError, clusto.get, None)

    def testDeleteEntity(self):

        e1 = Entity.query().filter_by(name=u'e1').one()

        d = Driver(e1)

        d.add_attr('deltest1', 'test')
        d.add_attr('deltest1', 'testA')



        clusto.delete_entity(e1)


        self.assertEqual([], clusto.get_entities(names=['e1']))

        self.assertEqual([], Driver.do_attr_query(key='deltest*', glob=True))



    def testDriverSearches(self):

        d = Driver('d1')

        self.assertRaises(NameError, clusto.get_driver_name, 'FAKEDRIVER')

        self.assertEqual(clusto.get_driver_name(Driver),
                         'entity')

        self.assertRaises(LookupError, clusto.get_driver_name, 123)

        self.assertEqual(clusto.get_driver_name('entity'),
                         'entity')

        self.assertEqual(clusto.get_driver_name(d.entity),
                         'entity')

    def testTypeSearches(self):

        d = Driver('d1')

        self.assertEqual(clusto.get_type_name('generic'),
                         'generic')

        self.assertEqual(clusto.get_type_name(d.entity),
                         'generic')

        self.assertRaises(LookupError, clusto.get_type_name, 123)


    def testAttributeOldVersionsInGetEntities(self):

        sl = [BasicServer('s' + str(x)) for x in range(10)]
        for n, s in enumerate(sl):
            s.add_attr(key='old', value="val")
            s.del_attrs(key='old')
            s.add_attr(key='new', value='foo')

        l=clusto.get_entities(attrs=[{'key':'old', 'value':'val'}])

        self.assertEqual(l, [])


    def testSiblings(self):

        d1 = Driver('d1')
        d2 = Driver('d2')
        d3 = Driver('d3')
        d4 = Driver('d4')
        d5 = Driver('d5')
        d6 = Driver('d6')
        d7 = Driver('d7')
        d8 = Driver('d8')

        db = Pool('db')
        web = Pool('web')
        dev = Pool('dev')
        prod = Pool('prod')
        alpha = Pool('alpha')
        beta = Pool('beta')

        db.set_attr('pooltype', 'role')
        web.set_attr('pooltype', 'role')

        db.insert(d1)
        db.insert(d2)
        db.insert(d3)
        db.insert(d7)
        db.insert(d8)

        web.insert(d4)
        web.insert(d5)
        web.insert(d6)
        web.insert(d7)

        map(prod.insert, [d1,d2,d4,d5])

        map(dev.insert, [d3,d6,d7,d8])

        map(alpha.insert, [d7, d8])
        map(beta.insert, [d3,d6])


        self.assertEquals(sorted([d2]),
                          sorted(d1.siblings()))

        self.assertEquals(sorted(d3.siblings()),
                          sorted([]))

        self.assertEquals(sorted(d3.siblings(parent_filter=lambda x: not x.attr_values('pooltype', 'role'),
                                             additional_pools=[web])),
                          sorted([d6]))

        self.assertEquals(sorted(d7.siblings(parent_filter=lambda x: not x.attr_values('pooltype', 'role'),
                                             additional_pools=[db])),
                          sorted([d8]))


    def testUnderscore(self):

        d1 = Driver('with_underscore')
        d2 = Driver('withZunderscore')

        self.assertEqual(clusto.get_by_name('with_underscore'),
                         d1)

        d1.add_attr('keyZfoo', 'bar')
        d1.add_attr('key_foo', 'baz')

        self.assertEqual(len(d1.attrs('key_foo')), 1)

        self.assertEqual(len(d1.attr_query('key_foo')), 1)




import clusto
from clusto.drivers import BasicRack, BasicServer
from clusto.test import testbase

class BasicRackTest(testbase.ClustoTestBase):

    def data(self):

        r1 = BasicRack('r1')
        r2 = BasicRack('r2')

        clusto.flush()

    def testAddingToRack(self):

        r1 = clusto.get_by_name('r1')

        s1 = BasicServer('s1')

        r1.insert(s1, 1)


        rt = clusto.get_by_name('r1')
        st = clusto.get_by_name('s1')

        self.assertEqual(len(r1.contents(subkey='ru')), 1)

        self.assertEqual(r1.contents(subkey='ru')[0].name, 's1')
        
        self.assertEqual(s1.parents(clusto_drivers=[BasicRack])[0].name, 'r1')

    def testMaxRackPosition(self):

        r1 = clusto.get_by_name('r1')

        self.assertRaises(TypeError, r1.insert, BasicServer('s1'), 400)

        self.assertRaises(TypeError, r1.insert, BasicServer('s2'), -13)

        clusto.flush()

    def testGettingThingInRack(self):

        r1 = clusto.get_by_name('r1')

        r1.insert(BasicServer('s1'), 40)

        clusto.flush()

        s1 = r1.get_device_in(40)

        self.assertEqual(s1.name, 's1')
        

    def testGettingRackAndU(self):

        r1, r2 = [clusto.get_by_name(r) for r in ['r1','r2']]

        s=BasicServer('s1')
        clusto.flush()
        r1.insert(s, 13)

        clusto.flush()

        s = clusto.get_by_name('s1')

        res = BasicRack.get_rack_and_u(s)

        
        self.assertEqual(res['rack'].name, 'r1')
        self.assertEqual(res['RU'], [13])

        res2 = BasicRack.get_rack_and_u(BasicServer('s2'))
        self.assertEqual(res2, None)

    def testCanOnlyAddToOneRack(self):
        """
        A device should only be able to get added to a single rack
        """

        
        r1, r2 = [clusto.get_by_name(r) for r in ['r1','r2']]

        s1 = BasicServer('s1')
        s2 = BasicServer('s2')
        
        r1.insert(s1, 13)
        self.assertRaises(Exception, r2.insert,s1, 1)
        
    def testCanAddADeviceToMultipleAdjacentUs(self):
        """
        you should be able to add a device to multiple adjacent RUs
        """

        r1, r2 = [clusto.get_by_name(r) for r in ['r1','r2']]

        s1 = BasicServer('s1')
        s2 = BasicServer('s2')
        
        r1.insert(s1, [1,2,3])

        clusto.flush()

        s = clusto.get_by_name('s1')

        self.assertEqual(sorted(BasicRack.get_rack_and_u(s)['RU']),
                         [1,2,3])

        self.assertRaises(TypeError, r1.insert, s2, [1,2,4])

    def testAddingToDoubleDigitLocationThenSingleDigitLocation(self):

        r1, r2 = [clusto.get_by_name(r) for r in ['r1','r2']]

        s1 = BasicServer('s1')
        s2 = BasicServer('s2')
        
        r1.insert(s1, 11)

        r1.insert(s2, 1)

        clusto.flush()

        s = clusto.get_by_name('s1')

        self.assertEqual(sorted(BasicRack.get_rack_and_u(s)['RU']),
                         [11])

        self.assertEqual(sorted(BasicRack.get_rack_and_u(s2)['RU']),
                         [1])




#!/usr/bin/env python
# -*- mode: python; sh-basic-offset: 4; indent-tabs-mode: nil; coding: utf-8 -*-
# vim: tabstop=4 softtabstop=4 expandtab shiftwidth=4 fileencoding=utf-8

import argparse
import operator
import random
import sys

import clusto
from clusto import drivers
from clusto import script_helper


class Allocate(script_helper.Script):
    '''
    Allocate servers from a given pool, you can specify filters
    to get a server with certain features. By default the script
    will search for the cheapest machine it can find
    '''

    def __init__(self):
        script_helper.Script.__init__(self)

    def _add_arguments(self, parser):
        parser.add_argument('--memory', '-m', type=int,
            help='Minimum system memory (in Gb)')
        parser.add_argument('--cores', '-c', type=int,
            help='Minimum number of system cores')
        parser.add_argument('--disk', '-d', type=int,
            help='Minimum disk size (in Gb)')
        parser.add_argument('--spindles', '-s', type=int,
            help='Minimum number of spindles')
        parser.add_argument('--to-pool', '-t', action='append', required=True,
            help='After allocating, add to the specified pools'
            ' (use more than once to set multi-pool)')
        parser.add_argument('--create-pools', '-C', action='store_true', default=False,
            help='Create the target pools if they do not exist')
        parser.add_argument('--from-pool', '-f',
            help='What pool to search in (you can set this in clusto.conf too '
                 'in allocate.pool: --from-pool > clusto.conf:allocate.pool > "unallocated")')
        parser.add_argument('--parent', '-p', required=True,
            help='Parent to start looking at. Can be a datacenter or a rack')
        parser.add_argument('number', nargs='?', type=int, default=1,
            help='How many machines to allocate, defaults to cheapest machine available')

    def __sort_servers(self, servers):
        tuples = []
        for server in servers:
            mem = server.attr_value(key='system', subkey='memory', default=None)
            cores = server.attr_value(key='system', subkey='cpucount',
                                      default=None)
            disk_size = server.attr_value(key='system', subkey='disk', default=None)
            disk_num = len(server.attrs(key='disk', subkey='size'))
            tuples.append((server, mem, cores, disk_size, disk_num))

        # do a single sort based on a four tuple as key!
        # tuple column order determines sort priority...
        tuples = sorted(tuples, key=lambda t:(t[3],t[2],t[1],t[4]))

        return [item[0] for item in tuples]

    def __make_filter(self, subkey, expected):
        def __filter(server):
            sys_info = server.attr_value(key='system', subkey=subkey, default=None)
            if not sys_info:
                return False
            if sys_info >= expected:
                return True
            else:
                return False
        
        def __filter_spindle(server):
            sys_len = len(server.attrs(key='disk', subkey='size'))
            return sys_len >= expected

        if subkey == 'spindles':
            return __filter_spindle
        else:
            return __filter

    def run(self, args):
        number = args.number

        if args.from_pool:
            self.debug('Grabbing pool from parameter')
            pool = args.from_pool
        else:
            self.debug('Grabbing pool from config file or default')
            pool = self.get_conf('allocate.pool', 'unallocated')
        self.debug('Unallocated pool is "%s"' % pool)

        pools = []
        try:
            if args.create_pools:
                pools = [ clusto.get_or_create(_, drivers.pool.Pool) for _ in args.pool ]
            else:
                pools = [ clusto.get_by_name(_, assert_driver=drivers.pool.Pool) for _ in args.to_pool ]
        except Exception as e:
            self.debug(e)
            self.error('There was an error when fetching/creating the pools to allocate the servers to')
            return 4
        self.debug('Target pool list: %s' % pools)

        try:
            pool = clusto.get_by_name(pool, assert_driver=drivers.pool.Pool)
        except Exception as e:
            self.debug(e)
            self.error('The pool "%s" does not exist' % pool)
            return 1

        try:
            parent = clusto.get(args.parent)
            if not parent:
                raise LookupError("Parent object is %s" % parent)
        except Exception as e:
            self.debug(e)
            self.error('The parent object "%s" does not exist' % args.parent)
            return 2
        parent = parent[0]
        if not isinstance(parent, drivers.racks.BasicRack) and not isinstance(parent, drivers.datacenters.BasicDatacenter):
            self.error('The parent "%s" is not a rack or a datacenter' % args.parent)
            return 2

        self.info('Searching for servers in "%s", this may take a while' % parent.name)

        unallocated = [ _ for _ in parent.contents(clusto_types=[drivers.servers.BasicServer], search_children=True) ]
        unallocated = [ _ for _ in unallocated if _ in pool and _.get_ips() ]
        self.debug('The unallocated list size is %d' % len(unallocated))
        if len(unallocated) < number:
            self.error('There are not enough servers in "%s" to fulfill your request' % args.parent)
            return 3

        filters = []
        if args.memory:
            filters.append(self.__make_filter('memory', args.memory * 1000))
        if args.disk:
            filters.append(self.__make_filter('disk', args.disk))
        if args.cores:
            filters.append(self.__make_filter('cpucount', args.cores))
        if args.spindles:
            filters.append(self.__make_filter('spindles', args.spindles))

        self.debug('Applying filters: %s' % filters)
        servers = []
        if not filters:
            servers = self.__sort_servers(unallocated)[:number]
        else:
            for func in filters:
                servers = filter(func, unallocated)[:number]
        self.debug('Server list: %s' % servers)

        for s in servers:
            pool.remove(s)
            for p in pools:
                p.insert(s)

        self.info('Allocated the following list of servers matching your filters '
            'were allocated from the pool "%s"' % pool.name)
        self.info('The servers were also added to the pools %s' % ','.join([ _.name for _ in pools ]))
        for s in servers:
            if s.get_ips():
                print s.get_ips()[0]
            else:
                print s.name

def main():
    allocate, args = script_helper.init_arguments(Allocate)
    return(allocate.run(args))


if __name__ == '__main__':
    sys.exit(main())




import clusto
from clusto.test import testbase 
import itertools

from clusto.drivers import *

from clusto.drivers.resourcemanagers.simplenamemanager import SimpleNameManagerException



class SimpleEntityNameManagerTests(testbase.ClustoTestBase):

    def data(self):

        n1 = SimpleEntityNameManager('foonamegen',
                                     basename='foo',
                                     digits=4,
                                     next=1,
                                     )


        n2 = SimpleEntityNameManager('barnamegen',
                                     basename='bar',
                                     digits=2,
                                     next=95,
                                     )
        
        clusto.flush()

    def testNamedDriverCreation(self):
        ngen = clusto.get_by_name('foonamegen')
        
        s1 = ngen.allocate(Driver)

        clusto.flush()

        d1 = clusto.get_by_name('foo0001')

        self.assertEquals(s1.name, d1.name)
        
    def testAllocateName(self):

        ngen = clusto.get_by_name('foonamegen')
        
        s1 = ngen.allocate(Driver)
        s2 = ngen.allocate(Driver)
        s3 = ngen.allocate(Driver)
        s4 = ngen.allocate(Driver)

        clusto.flush()

        self.assertEqual(s1.name, 'foo0001')
        self.assertEqual(s2.name, 'foo0002')
        self.assertEqual(s3.name, 'foo0003')
        self.assertEqual(s4.name, 'foo0004')

    def testNoLeadingZeros(self):

        ngen = clusto.get_by_name('barnamegen')

        s1 = ngen.allocate(Driver)
        s2 = ngen.allocate(Driver)
        s3 = ngen.allocate(Driver)
        s4 = ngen.allocate(Driver)

        clusto.flush()

        self.assertEqual(s1.name, 'bar95')
        self.assertEqual(s2.name, 'bar96')
        self.assertEqual(s3.name, 'bar97')
        self.assertEqual(s4.name, 'bar98')

    def testTooManyDigits(self):
        
        ngen = clusto.get_by_name('barnamegen')

        s1 = ngen.allocate(Driver)
        s2 = ngen.allocate(Driver)
        s3 = ngen.allocate(Driver)
        s4 = ngen.allocate(Driver)

        s5 = ngen.allocate(Driver)
        self.assertRaises(SimpleNameManagerException,
                          ngen.allocate, Driver)


    def testAllocateManyNames(self):
        
        ngen = clusto.get_by_name('foonamegen')

        for i in xrange(50):
            ngen.allocate(Driver)

        self.assertRaises(LookupError, clusto.get_by_name, 'foo0051')
        self.assertEqual(clusto.get_by_name('foo0050').name, 'foo0050')


    def testAllocateGivenName(self):

        ngen = clusto.get_by_name('foonamegen')

        d = ngen.allocate(Driver, 'testname')

        self.assertEqual(d.name, 'testname')

class SimpleNameManagerTests(testbase.ClustoTestBase):

    def data(self):
        n1 = SimpleNameManager('foonamegen',
                               basename='foo',
                               digits=4,
                               startingnum=1,
                               )

        clusto.flush()

    def testAllocateManyNames(self):
        
        ngen = clusto.get_by_name('foonamegen')

        d = Driver('foo')

        for i in xrange(50):
            ngen.allocate(d)
            
        
        self.assertEqual(len(SimpleNameManager.resources(d)), 50)



#!/usr/bin/env python
# -*- mode: python; sh-basic-offset: 4; indent-tabs-mode: nil; coding: utf-8 -*-
# vim: tabstop=4 softtabstop=4 expandtab shiftwidth=4 fileencoding=utf-8

import argparse
import re
import sys

import clusto
from clusto.drivers import IPManager
from clusto import script_helper


class InitDB(script_helper.Script):
    '''
    Initialize the clusto database
    '''

    def __init__(self):
        script_helper.Script.__init__(self)

    def run(self, args):
        try:
            clusto.init_clusto()
        except Exception, e:
            print 'Error creating clusto database: %s' % str(e)
            return -1

def main():
    initdb, args = script_helper.init_arguments(InitDB)
    return(initdb.run(args))

if __name__ == '__main__':
    sys.exit(main())




import clusto
from clusto.test import testbase 
import itertools

from clusto.drivers import *

from clusto.drivers.resourcemanagers.simplenummanager import *



class SimpleNumManagerTests(testbase.ClustoTestBase):

    def data(self):

        n1 = SimpleNumManager('numgen1', next=1)

        n2 = SimpleNumManager('numgen2', maxnum=4, next=0)
        
        clusto.flush()

    def testAllocateNum(self):

        ngen = clusto.get_by_name('numgen1')
        
        d = Driver('foo')
        s1 = ngen.allocate(d)
        s2 = ngen.allocate(d)
        s3 = ngen.allocate(d)
        s4 = ngen.allocate(d)

        self.assertEqual(ngen.owners(1), [d])
        self.assertEqual(ngen.owners(2), [d])
        self.assertEqual(ngen.owners(3), [d])
        self.assertEqual(ngen.owners(4), [d])


    def testAllocateMaxNum(self):
        
        d = Driver('foo')

        ngen = clusto.get_by_name('numgen2')

        s1 = ngen.allocate(d)
        s1 = ngen.allocate(d)
        s1 = ngen.allocate(d)
        s1 = ngen.allocate(d)
        s1 = ngen.allocate(d)

        self.assertRaises(SimpleNumManagerException, ngen.allocate, d)
        
        



"""
Test the basic Entity object
"""

import unittest
from clusto.test import testbase
import datetime

import clusto
from clusto.schema import *

class TestEntitySchema(testbase.ClustoTestBase):

    def testCreateEntityObject(self):

        curver = clusto.get_latest_version_number()

        e1 = Entity('e1')
        e2 = Entity('e2')

        res = Entity.query().filter_by(name=u'e1')

        self.assertEqual(res.count(),1)

        e = res.all()[0]

        self.assertEqual(e.name, 'e1')


    def testOutputEntityObject(self):

        expectedout = "e1"
        
        e1 = Entity('e1')

        self.assertEqual(str(e1), expectedout)

        clusto.flush()

        self.assertEqual(str(Entity.query().filter_by(name=u'e1')[0]), expectedout)
        

    def testDeleteEntity(self):

        e1 = Entity('e1')

        clusto.flush()

        self.assertEqual(Entity.query().filter_by(type='entity').count(), 1)

        e1.delete()

        clusto.flush()

        self.assertEqual(Entity.query().filter_by(type='entity').count(), 0)
    
class TestEntityAttributes(testbase.ClustoTestBase):

    def data(self):

        Entity('e1')
        Entity('e2')
        Entity('e3')

        clusto.flush()

    def testAddingAttribute(self):

        e = Entity.query().filter_by(name=u'e2').one()

        e1 = Entity.query().filter_by(name=u'e1').one()

                
        self.assertEqual(e.name, 'e2')

        e.add_attr(key='one', value=1)
        e.add_attr(key='two', value=2)

        clusto.flush()

        q = Attribute.query().filter_by(entity_id=e.entity_id,
                                               key=u'two').one() 
        self.assertEqual(q.value, 2)

        q = Attribute.query().filter_by(entity_id=e.entity_id,
                                               key=u'one').one()

        self.assertEqual(q.value, 1)
        

    def testAddingDateAttribute(self):

        e1 = Entity.query().filter_by(name=u'e1').one()

        d = datetime.datetime(2007,12,16,7,46)
        
        e1.add_attr('somedate', d)

        clusto.flush()

        q = Attribute.query().filter_by(entity_id=e1.entity_id,
                                               key=u'somedate').one()

        self.assertEqual(q.value, d)
        
    def testData(self):

        q = Entity.query().\
               filter(not_(Entity.type=='clustometa')).count()

        self.assertEqual(q, 3)
        
    def testEmptyAttributes(self):
        """
        If I set no attributes there shouldn't be any in the DB except the
        clusto meta attributes
        """
        
        q = Attribute.query().join('entity').\
               filter(not_(Entity.type=='clustometa')).count()

        self.assertEqual(q, 0)
        
    def testRelationAttribute(self):

        e1 = Entity.query().filter_by(name=u'e1').one()
        
        e4 = Entity('e4')
        e4.add_attr(key='e1', value=e1)
        
        clusto.flush()


        e4 = Entity.query().filter_by(name=u'e4').one()

        attr = e4.attrs[0]

        self.assertEqual(attr.relation_value, e1)

    def testStringAttribute(self):

        e2 = Entity.query().filter_by(name=u'e2').one()

        e2.add_attr(key='somestring', value='thestring')

        clusto.flush()

        q = Attribute.query().filter_by(entity=e2,
                                               key=u'somestring').one()

        self.assertEqual(q.value, u'thestring')

    def testJSONAttribute(self):

        e2 = Entity.query().filter_by(name=u'e2').one()

        e2.add_attr(key="somejson", value=['foo', 'bar'])

        clusto.flush()

        q = Attribute.query().filter_by(entity=e2,
                                        key=u'somejson').one()

        self.assertEqual(q.value, ['foo', 'bar'])

        e2.add_attr(key="otherjson", value={'foo':200, 'bar':'test'})

        q = Attribute.query().filter_by(entity=e2,
                                        key=u"otherjson").one()

        self.assertEqual(q.value, {'foo':200, 'bar':'test'})

    def testIntAttribute(self):

        e4 = Entity('e4')
        e4.add_attr(key='someint', value=10)

        clusto.flush()

        q = Attribute.query().filter_by(entity=e4,
                                               key=u'someint').one()

        self.assertEqual(q.value, 10)

    def testMultipleAttributes(self):

        e2 = Entity.query().filter_by(name=u'e2').one()

        e2.add_attr(key='somestring', number=1, subkey='foo',
                                   value='thestring')

        e2.add_attr(key='somestring', number=1, subkey='foo',
                                   value='thestring')


        clusto.flush()

        q = Attribute.query().filter_by(entity=e2,
                                               key=u'somestring').all()

        self.assertEqual([a.value for a in q], 
                         ['thestring', 'thestring'])

    def testEntityDeleteRelations(self):

        e1 = Entity.query().filter_by(name=u'e1').one()
        e2 = Entity.query().filter_by(name=u'e2').one()

        e1.add_attr('pointer1', e2)

        clusto.flush()

        self.assertEqual(Entity.query().\
                            filter_by(type='entity').count(),
                         3)

        self.assertEqual(Attribute.query().\
                            filter(and_(Entity.entity_id==Attribute.entity_id,
                                        Entity.type=='entity',
                                        
                                        )).count()
                         , 1)

        e2new = Entity.query().filter_by(name=u'e2').one()

        e2new.delete()

        self.assertEqual(Entity.query().\
                            filter_by(type='entity').count(),
                         2)
        
        self.assertEqual(Attribute.query().\
                            filter(and_(Entity.entity_id==Attribute.entity_id,
                                        Entity.type=='entity')).count(),
                         0)

        clusto.flush()

        self.assertEqual(Entity.query().\
                            filter_by(type='entity').count(),
                         2)
        self.assertEqual(Attribute.query().\
                            filter(and_(Entity.entity_id==Attribute.entity_id,
                                        Entity.type=='entity')).count(),
                         0)


    def testAccessRelationAttributesMultipleTimes(self):
        e1 = Entity.query().filter_by(name=u'e1').one()
        e2 = Entity.query().filter_by(name=u'e2').one()

        e1.add_attr('foo', 2)
        e1.add_attr('foo', e2)

        clusto.flush()
        e1 = Entity.query().filter_by(name=u'e1').one()
        self.assertEqual(len(list(e1.attrs)), 2)
        self.assertEqual(len(list(e1.attrs)), 2)
        self.assertEqual(len(list(e1.attrs)), 2)

        
    
class TestEntityReferences(testbase.ClustoTestBase):

    def data(self):
        
        e1 = Entity('e1')
        e2 = Entity('e2')
        e3 = Entity('e3')

        e3.add_attr(key='e1', value=e1)
        e3.add_attr(key='e2', value=e2)

        clusto.flush()
    
    def testReference(self):

        e1 = Entity.query().filter_by(name=u'e1').one()
        e2 = Entity.query().filter_by(name=u'e2').one()
        e3 = Entity.query().filter_by(name=u'e3').one()

        self.assertEqual(e1.references[0].entity,
                         e2.references[0].entity)

        self.assertEqual(e3,
                         e2.references[0].entity)

    def testReferenceDelete(self):

        e1 = Entity.query().filter_by(name=u'e1').one()


        e3 = Entity.query().filter_by(name=u'e3').one()

        
        e3.delete()

        self.assertEqual(len(list(e1.references)), 0)

        clusto.flush()

        e1a = Entity.query().filter_by(name=u'e1').one()

        self.assertEqual(len(list(e1a.references)), 0)
        self.assertEqual(id(e1a), id(e1))

        e2 = Entity.query().filter_by(name=u'e2').one()

        self.assertEqual(len(list(e2.references)), 0)





TYPELIST = {}
DRIVERLIST = {}
RESERVEDATTRS = {}

class ClustoDriver(type):
    """
    Metaclass for all clusto drivers
    """
    def __init__(cls, name, bases, dct):

        if not hasattr(cls, '_driver_name'):
            raise DriverException("Driver %s missing _driver_name attribute"
                                  % cls.__name__)

        if cls._driver_name in DRIVERLIST:
            raise KeyError("class '%s' is trying to add the driver_name '%s' "
                           "to the driver list but that name is already "
                           "claimed by the '%s' class."
                           % (cls.__name__,
                              cls._driver_name,
                              DRIVERLIST[cls._driver_name].__name__))
        

        DRIVERLIST[cls._driver_name] = cls
        TYPELIST[cls._clusto_type] = cls

        # setup properties
        if not isinstance(cls._properties, dict):
            raise TypeError('_properties of %s is not a dict type.',
                            cls.__name__)
        

        super(ClustoDriver, cls).__init__(name, bases, dct)




from clusto.drivers.base import Device
from clusto.drivers.devices.common import PortMixin, IPMixin

class BasicServer(IPMixin, PortMixin, Device):
    """
    server
    """

    _clusto_type = 'server'
    _driver_name = 'basicserver'

    _properties = {'model':None,
                   'manufacturer':None}

    _portmeta = {'pwr-nema-5': {'numports':1},
                 'nic-eth': {'numports':2},
                 'console-serial' : { 'numports':1, }
                 }



class BasicVirtualServer(BasicServer, IPMixin):

    _clusto_type = 'server'
    _driver_name = 'basicvirtualserver'

    def power_on(self, *args, **kwargs):
        raise NotImplementedError()

    def power_off(self, *args, **kwargs):
        raise NotImplementedError()

    def power_reboot(self, *args, **kwargs):
        raise NotImplementedError()

    def create(self, pool, **kwargs):
        raise NotImplementedError()

    def destroy(self):
        raise NotImplementedError()




from clusto.drivers.base.clustodriver import *
from clusto.drivers.base.driver import *
from clusto.drivers.base.clustometa import *
from clusto.drivers.base.device import *
from clusto.drivers.base.location import *
from clusto.drivers.base.resourcemanager import *
from clusto.drivers.base.controller import *




import os.path
import logging.handlers
import logging
import sys
import os

try:
    import simplejson as json
except ImportError:
    import json

LEVELS = {
    'DEBUG':    logging.DEBUG,
    'INFO':     logging.INFO,
    'WARNING':  logging.WARNING,
    'ERROR':    logging.ERROR,
    'CRITICAL': logging.CRITICAL,
}

files = [
    '/etc/clusto/services.conf',
    '%s/.clusto/services.conf' % os.environ.get('HOME', '/tmp'),
    'services.conf',
]

config = None
for filename in files:
    if os.path.exists(filename):
        try:
            config = json.load(file(filename, 'r'))
            break
        except:
            sys.stderr.write('Unable to parse config file %s: %s\n' % (filename, sys.exc_info()[1]))

if not config:
    sys.stderr.write('Unable to find services.conf!\n')

def conf(key, **kwargs):
    obj = config
    for k in key.split('.'):
        try:
            obj = obj[k]
        except KeyError, ke:
            if 'default' in kwargs.keys():
                return kwargs['default']
            else:
                raise KeyError(ke)
    return obj

def get_logger(name, level='INFO'):
    log = logging.getLogger(name)

    fmt = logging.Formatter('%(asctime)s %(name)s %(levelname)s %(message)s', '%Y-%m-%d %H:%M:%S')
    stdout = logging.StreamHandler()
    stdout.setFormatter(fmt)

    fmt = logging.Formatter('%(name)s %(message)s')
    syslog = logging.handlers.SysLogHandler()
    syslog.setFormatter(fmt)

    log.addHandler(stdout)
    log.addHandler(syslog)
    log.setLevel(LEVELS[level])
    return log



#!/usr/bin/env python
# -*- mode: python; sh-basic-offset: 4; indent-tabs-mode: nil; coding: utf-8 -*-
# vim: tabstop=4 softtabstop=4 expandtab shiftwidth=4 fileencoding=utf-8
#
# Base script helper
# Copyright 2011, Jorge A Gallegos <kad@blegh.net>

import argparse
import ConfigParser
import glob
import logging
import os
import sys

import clusto


levels = {
    'INFO': logging.INFO,
    'DEBUG': logging.DEBUG,
    'ERROR': logging.ERROR,
    'FATAL': logging.FATAL,
    'WARN': logging.WARN,
}


class CmdLineError(Exception):
    pass


class Script(object):
    '''
    Base script helper class
    '''

    logger = None
    config = None

    def __init__(self):
        pass

    def _get_description(self):
        '''
        Returns the docstring for the class if there's one,
        otherwise it returns a generic blurb
        '''

        if self.__class__.__doc__:
            help_string = self.__class__.__doc__
        else:
            help_string = '%s command help' % self.__class__.__name__.capitalize()
        return help_string

    def _setup_subparser(self, subparsers):
        '''
        Sets up the command subparser, this should (in theory at least)
        only be called from within add_subparser()
        '''

        command_name = self.__module__.split('.')[-1].lower().replace('_', '-')
        parser = subparsers.add_parser(command_name, help=self._get_description())
        return parser

    def add_subparser(self, subparser):
        '''
        Adds the command subparser to the base parser. This method should
        always call _setup_subparser so the command becomes available in
        the argparse namespace, otherwise it won't be listed
        '''

        parser = self._setup_subparser(subparser)
        if '_add_arguments' in dir(self) and \
                            callable(getattr(self, '_add_arguments')):
            self._add_arguments(parser)

    def set_logger(self, logger):
        '''
        Sets the class logger to whatever logging facility we provide
        '''
        self.logger = logger

    def log(self, msg, level=logging.INFO):
        '''
        I dislike having to type self.logger.log()
        '''
        self.logger.log(level=level, msg=msg)

    def error(self, msg):
        '''
        Passthrough to self.logger.error()
        '''
        self.log(msg, level=logging.ERROR)

    def warn(self, msg):
        '''
        Passthrough to self.logger.warn()
        '''
        self.log(msg, level=logging.WARN)

    def fatal(self, msg):
        '''
        Passthrough to self.logger.fatal()
        '''
        self.log(msg, level=logging.FATAL)

    def debug(self, msg):
        '''
        Passthrough to self.logger.debug()
        '''
        self.log(msg, level=logging.DEBUG)

    def info(self, msg):
        '''
        Passthrough to self.logger.info()
        '''
        self.log(msg)

    def run(self, *args, **kwargs):
        '''
        Main command loop, should be implemented in child classes
        '''
        raise NotImplementedError()

    def get_conf(self, path, default=None):
        '''
        Returns the config value
        '''

        (section, option) = path.split('.')
        if self.config.has_option(section, option):
            return self.config.get(section, option)
        else:
            return default

    def init_script(self, args, logger=None):
        '''
        Initialize the clusto environment for clusto scripts.

        Connects to the clusto database, returns a python SafeConfigParser
        '''

        if logger:
            self.set_logger(logger)
        if 'CLUSTOCONFIG' in os.environ:
            filename = os.environ['CLUSTOCONFIG']
        else:
            filename = args.config
        self.config = load_config(filename, args.dsn, logger)
        self.debug('Connecting to %s' % self.config.get('clusto', 'dsn'))
        clusto.connect(self.config)

        return self.config

def load_plugins(config):
    plugins = {} #{'clusto': clusto}
    if config.has_option('clusto', 'plugins'):
        for plugin in config.get('clusto', 'plugins').split(','):
            plugins[plugin] = None
    if 'CLUSTOPLUGINS' in os.environ:
        for plugin in os.environ['CLUSTOPLUGINS'].split(','):
            plugins[plugin] = None
    for plugin in plugins.keys():
        module = __import__(plugin)
        plugins[plugin] = module
    return plugins

def load_config(filename=None, dsn=None, logger=None):
    '''
    Find, parse, and return the configuration data needed by clusto
    '''

    if not logger:
        logger = get_logger()

    config = ConfigParser.SafeConfigParser()
    config.add_section('clusto')

    if filename is not None:
        if not os.path.exists(os.path.realpath(filename)):
            msg = "Config file %s doesn't exist." % filename
            logger.error(msg)
            raise CmdLineError(msg)
        
        logger.debug('Reading %s' % filename)
        config.read([filename])

        if not config.has_section('clusto'):
            config.add_section('clusto')

        if config.has_option('clusto', 'include'):
            for pattern in config.get('clusto', 'include').split():
                for filename in glob.glob(pattern):
                    fn = os.path.realpath(filename)
                    logger.debug('Trying to add config file: "%s"' % fn)
                    try:
                        config.read(fn)
                    except ConfigParser.ParsingError, cppe:
                        logger.warn(cppe)
                    except Exception, e:
                        logger.debug(e)

    if dsn:
        config.set('clusto', 'dsn', dsn)
    elif 'CLUSTODSN' in os.environ:
        config.set('clusto', 'dsn', os.environ['CLUSTODSN'])

    if not config.has_option('clusto', 'dsn'):
        raise CmdLineError("No database given for clusto data.")

    logger.debug('Loaded plugins: %s' % (load_plugins(config),))

    return config

def demodule(module):
    '''
    Returns a class out of a given module name by doing:
    some_name_in_this_form => SomeNameInThisForm()
    '''

    klass = ''.join([_.capitalize() for _ in module.split('_')])
    module = __import__('clusto.commands.%s' % module, fromlist=['clusto.commands'])
    klass = getattr(module, klass)
    return klass

def setup_base_parser(add_help=False):
    '''
    Setups the base parser with common options, this is used here down below
    or can alternatively be used in the commands main() call for standalone
    script support
    '''

    import argparse
    parser = argparse.ArgumentParser(prog=os.path.basename(sys.argv[0]),
        description='Clusto script', conflict_handler='resolve',
        add_help=add_help, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('--config', default='/etc/clusto/clusto.conf',
        help='Config file (defaults to /etc/clusto/clusto.conf)')
    parser.add_argument('--loglevel', default='INFO',
        help='Sets the logging level (defaults to INFO)')
    parser.add_argument('--dsn', help='Alternate DSN to use')
    return parser

def get_logger(name='', loglevel='INFO'):
    '''
    Returns a basic std{out,err} logger
    '''

    format='%(name)s: %(levelname)-8s %(message)s'
    logging.basicConfig(format=format)
    log = logging.getLogger('command.%s' % name)
    log.setLevel(levels[loglevel])
    return log

def init_arguments(klass, formatter_class=argparse.RawTextHelpFormatter):
    obj = klass()
    parent_parser = setup_base_parser()
    this_parser = argparse.ArgumentParser(parents=[parent_parser],
        description=obj._get_description(), formatter_class=formatter_class)
    if '_add_arguments' in dir(obj) and callable(getattr(obj, '_add_arguments')):
        obj._add_arguments(this_parser)
    args = this_parser.parse_args()
    log = get_logger(klass.__module__.split('.')[-1], args.loglevel)
    obj.init_script(args=args, logger=log)
    return obj, args

def main():
    # should only be imported if called from the cli
    from clusto import commands

    parser = setup_base_parser(add_help=True)
    subparsers = parser.add_subparsers(title='Available clusto commands', dest='subparser_name')
    help_subparser = subparsers.add_parser('help', help='Print clusto help')
    for dirpath, dirname, filenames in os.walk(commands.__path__[0]):
        for fn in filenames:
            if not fn.endswith('pyc') and not fn.startswith('__'):
                module = fn.split('.')[0]
                try:
                    klass = demodule(module)
                    klass = klass()
                    klass.add_subparser(subparsers)
                    klass = None
                except Exception, e:
                    pass
    args = parser.parse_args()

    log = get_logger(args.subparser_name, args.loglevel)
    if len(sys.argv) < 2:
        parser.print_help()
        return 0
    if args.subparser_name == 'help':
        parser.print_help()
        return 0
    log.debug('Loading from clusto frontend: %s' % args.subparser_name)
    klass = demodule(args.subparser_name.replace('-', '_'))
    klass = klass()
    klass.set_logger(log)
    try:
        klass.init_script(args=args, logger=log)
        return(klass.run(args))
    except Exception as e:
        log.error(str(e))
        return 99


if __name__ == '__main__':
    sys.exit(main())




#!/usr/bin/env python
# -*- mode: python; sh-basic-offset: 4; indent-tabs-mode: nil; coding: utf-8 -*-
# vim: tabstop=4 softtabstop=4 expandtab shiftwidth=4 fileencoding=utf-8

import argparse
import sys

import clusto
from clusto import drivers
from clusto import script_helper


class Deallocate(script_helper.Script):
    '''
    This is a script that will deallocate machines from pool. That is, it will
    cleanup most attributes (except for system attributes) and will remove it
    from every pool it is on to then put it in a pre-defined 'unallocated' pool
    '''

    keep_attrs = ['port-console-serial', 'port-nic-eth', 'ip', 'system',
        'port-pwr-nema-5', 'memory', 'disk', 'processor']
    hosts = []

    def __init__(self):
        script_helper.Script.__init__(self)

    def _add_arguments(self, parser):
        parser.add_argument('--pool',
            help='Target pool after deallocation (you can set this in clusto.conf too '
                 'in deallocate.pool: --pool > clusto.conf:deallocate.pool > "unallocated")')
        parser.add_argument('--keep-attrs',
            help='Comma-separated keys to keep additional to whatever is configured '
                 'in clusto.conf in deallocate.keep_attrs')
        parser.add_argument('--reboot', action='store_true', default=False,
            help='Reboot the server after deallocating (defaults to False)')
        parser.add_argument('--shutdown', action='store_true', default=False,
            help='Power-off the server after deallocating (defaults to False)')
        parser.add_argument('--force-yes', action='store_true',
            default=False, help='Force yes to deallocate')
        parser.add_argument('objects', nargs='+',
            help='Server names or IP addresses')

    def _clean_attrs(self, host):
        '''
        Cleans the attributes from a given host, except for
        those in the "to keep" list
        '''
#   Clean attributes
        self.info('Cleaning up attributes from "%s"' % host.name)
        attrs = set(sorted([x.key for x in host.attrs() if x.key not in self.keep_attrs]))
        self.debug("List of attrs to remove: %s" % attrs)
        for attr in attrs:
            host.del_attrs(key=attr)

    def _deallocate(self, host):
        '''
        Remove the given host from all the parent pools
        '''
        self.info('Deallocating %s...' % host.name)
        pools = host.parents(clusto_types=[drivers.pool.Pool])
#   Remove from pools
        for pool in pools:
            if pool.name != 'fai' and pool.name != 'unallocated':
                self.info('Removing "%s" from pool "%s"' % (host.name, pool.name))
                pool.remove(host)

    def run(self, args):
#       get the 'unallocated' pool
        if args.pool:
            self.debug('Grabbing pool from parameter')
            pool = args.pool
        else:
            self.debug('Grabbing pool from config file or default')
            pool = self.get_conf('deallocate.pool', 'unallocated')
        self.debug('Unallocated pool is "%s"' % pool)

#       Load the attributes from the conf file
        for attr in self.get_conf('deallocate.keep_attrs', '').split(','):
            if attr.split():
                self.keep_attrs.append(attr.strip())
#       Append attributes from the command line, if any
        if args.keep_attrs:
            for attr in args.keep_attrs.split(','):
                if attr not in self.keep_attrs:
                    self.keep_attrs.append(attr)
        self.debug('Final list of attrs to keep: %s' % self.keep_attrs)
        for obj in args.objects:
            obj = clusto.get(obj)
            if obj not in self.hosts:
                self.hosts.extend(obj)
        self.info('This is the list of servers/ipaddresses that you will deallocate')
        for host in self.hosts:
            if host.get_ips():
                print '%s >> %s' % (host.name, ' '.join(host.get_ips()))
            else:
                print host.name

#       Ask for confirmation
        if not args.force_yes:
            sys.stdout.write('Are you absolutely sure you want to continue %s (yes/No)? ')
            sys.stdout.flush()

            try:
                line = sys.stdin.readline().strip()
            except KeyboardInterrupt:
                line = False

            if line != 'yes':
                print 'Aborting'
                return 7

        for host in self.hosts:
            if not isinstance(host, drivers.servers.BasicServer):
                self.warn('Cannot deallocate "%s" because is not a server' % host.name)
                continue
            self._deallocate(host)
            self._clean_attrs(host)
            if host not in pool:
                pool.insert(host)
            if args.reboot:
                host.power_reboot(captcha=False)
                continue
            if args.shutdown:
                host.power_off(captcha=False)
                continue
        self.info('Done.')

def main():
    deallocate, args = script_helper.init_arguments(Deallocate)
    return(deallocate.run(args))

if __name__ == '__main__':
    sys.exit(main())





if __name__ == '__main__':

    import os
    import sys


    sys.path.append(os.path.realpath('.'))


import unittest
import clusto.test


def gettests(tests=None):
    if not tests:
        tests = ('clusto.test.base', 'clusto.test.drivers',
                 'clusto.test.usage',)

    suite = unittest.defaultTestLoader.loadTestsFromNames(tests)

    return suite


def runtests(tests=None, db='sqlite:///:memory:', echo=False):

    clusto.test.testbase.DB=db
    clusto.test.testbase.ECHO=echo
    suite = gettests(tests)
    runner = unittest.TextTestRunner()    
    runner.run(suite)




if __name__ == '__main__':

    import optparse

    parser = optparse.OptionParser()
    parser.add_option('--db', dest='dsn', 
                      help='specifies which db to test against',
                      default='sqlite:///:memory:')
    parser.add_option('--echo', dest='echo', action='store_true', default=False,
                      help="Echo sqlalchemy sql")
    
    (options, args) = parser.parse_args()
    runtests(args, options.dsn, options.echo)




import clusto
from clusto.test import testbase 
import itertools

from clusto.drivers import *
from clusto.exceptions import PoolException

class PoolTests(testbase.ClustoTestBase):

    def data(self):

        d1 = Driver('d1')
        d2 = Driver('d2')

        p1 = Pool('p1')

        clusto.flush()


    def testPoolCreate(self):

        p3 = Pool('p3')
        d3 = Driver('d3')
        
        clusto.flush()

        q = clusto.get_by_name('p3')

        self.assertTrue(isinstance(q, Pool))

        self.assertFalse(isinstance(clusto.get_by_name('d3'), Pool))

    def testPoolMembers(self):

        d1, d2, p1 = map(clusto.get_by_name, ('d1', 'd2', 'p1'))
        
        p1.insert(d1)
        p1.insert(d2)
        
        clusto.flush()


        q = clusto.get_by_name('p1')

        membernames = sorted([x.name for x in p1.contents()])

        self.assertEqual(membernames, sorted(['d1','d2']))

    def testGetPools(self):

        d1, d2, p1 = [clusto.get_by_name(i) for i in ['d1', 'd2', 'p1']]

        p2 = Pool('p2')
        
        p1.insert(d1)
        p2.insert(d1)
        p1.insert(d2)


        self.assertEqual(sorted(Pool.get_pools(d1)),
                         sorted([p1,p2]))

    def testGetPoolsMultiLevel(self):

        d1, d2, p1 = [clusto.get_by_name(i) for i in ['d1', 'd2', 'p1']]

        p2 = Pool('p2')
        p3 = Pool('p3')
        p4 = Pool('p4')
        d3 = Driver('d3')
        
        p1.insert(d1)
        p2.insert(d1)
        p1.insert(d2)

        p1.insert(p3)
        p3.insert(p4)
        p4.insert(d3)

        self.assertEqual(sorted(Pool.get_pools(d1, allPools=True)),
                         sorted([p1,p2]))

                
        self.assertEqual(sorted(set(Pool.get_pools(d3, allPools=True))),
                         sorted([p1, p3, p4]))

    def testPoolsIterator(self):

        
        A = Pool('A')

        d1, d2 = [clusto.get_by_name(i) for i in ['d1', 'd2']]

        B = Pool('B')
        C = Pool('C')
        A1 = Pool('A1')
        B2 = Pool('B2')
        B1 = Pool('B1')
        C1 = Pool('C1')

        C1.insert(C)
        B1.insert(B)
        A1.insert(B)

        A1.insert(A)
        B2.insert(A)

        A.insert(d1)
        B.insert(d1)
        C.insert(d1)

        clusto.flush()

        self.assertEqual([x.name for x in Pool.get_pools(d1)],
                         [u'A', u'B', u'C', u'A1', u'B2', u'B1', u'A1', u'C1'])

        self.assertEqual([x.name for x in Pool.get_pools(d1, allPools=False)],
                         [u'A', u'B', u'C'])


    def testPoolAttrs(self):

        d1, d2, p1 = map(clusto.get_by_name, ('d1', 'd2', 'p1'))

        p1.add_attr('t1', 1)
        p1.add_attr('t2', 2)

        d1.add_attr('t3', 3)

        p1.insert(d1)
        p1.insert(d2)
        

        clusto.flush()

        d2 = clusto.get_by_name('d2')

        self.assertEqual(sorted(d2.attrs(merge_container_attrs=True)), sorted(p1.attrs()))


        self.assertEqual(sorted(['t1', 't2', 't3']),
                         sorted([x.key for x in d1.attrs(merge_container_attrs=True)]))


    def testPoolAttrsOverride(self):

        d1, d2, p1 = map(clusto.get_by_name, ('d1', 'd2', 'p1'))

        p1.add_attr('t1', 1)
        p1.add_attr('t2', 2)

        p1.insert(d1)
        d1.add_attr('t1', 'foo')
        clusto.flush()

        
    def testFilterPoolContents(self):
        
        d1, d2, p1 = map(clusto.get_by_name, ('d1', 'd2', 'p1'))

        p1.insert(d1)
        p1.insert(d2)
        p1.insert(BasicServer('s1'))
        p1.insert(BasicServer('s2'))
        p1.insert(BasicNetworkSwitch('sw1'))

        self.assertEqual(p1.contents(clusto_types=[BasicNetworkSwitch]),
                         [clusto.get_by_name('sw1')])

    def testAddToPoolMultipleTimes(self):
        
        d1, d2, p1 = map(clusto.get_by_name, ('d1', 'd2', 'p1'))

        p1.insert(d1)

        self.assertRaises(PoolException, p1.insert, d1)
        

    def testDelAttrs(self):

        d1, d2, p1 = map(clusto.get_by_name, ('d1', 'd2', 'p1'))

        p1.insert(d1)
        p1.insert(d2)

        p1.add_attr('testkey', 'foo')
        
        self.assertEqual(sorted(p1.contents()),
                         sorted([d1,d2]))

        self.assertEqual([(a.key, a.value) for a in p1.attrs()],
                         [('testkey','foo')])
        
        p1.del_attrs()

        self.assertEqual(sorted(p1.contents()),
                         sorted([d1,d2]))

        self.assertEqual(p1.attrs(), [])


    def testGetFromPools(self):
        p1 = clusto.get_or_create('p1', Pool)
        p2 = clusto.get_or_create('p2', Pool)

        s1 = BasicServer('s1')
        s2 = BasicServer('s2')
        s3 = BasicServer('s3')
        s4 = BasicServer('s4')
        s5 = BasicServer('s5')
        s6 = BasicServer('s6')

        p1.insert(s1)
        p1.insert(s2)
        p1.insert(s3)
        p2.insert(s3)
        p2.insert(s4)
        p2.insert(s5)

        self.assertEqual(sorted([s1, s2, s3]),
                         sorted(clusto.get_from_pools(pools=[p1],
                                                      clusto_types=[BasicServer])))
        self.assertEqual(sorted([s3, s4, s5]),
                         sorted(clusto.get_from_pools(pools=[p2],
                                                      clusto_types=[BasicServer])))
        self.assertEqual(sorted([s3]),
                         sorted(clusto.get_from_pools(pools=[p1, p2],
                                                      clusto_types=[BasicServer])))

        self.assertRaises(LookupError,
                          clusto.get_from_pools, 'p1', 'non-existent-pool')




from clusto.schema import *
from clusto.exceptions import *

from clusto.drivers import DRIVERLIST, TYPELIST, Driver, ClustoMeta, IPManager
from sqlalchemy.exc import InvalidRequestError
from sqlalchemy import create_engine
from sqlalchemy.pool import SingletonThreadPool

from clusto import drivers

import threading
import logging
import time
import re

driverlist = DRIVERLIST
typelist = TYPELIST

def connect(config, echo=False):
    """Connect to a given Clusto datastore.

    Accepts a config object with (at least) a DSN string

    e.g. mysql://user:pass@example.com/clustodb
    e.g. sqlite:///somefile.db

    @param config: the config object
    """

    SESSION.configure(bind=create_engine(config.get('clusto', 'dsn'),
                                         echo=echo,
                                         poolclass=SingletonThreadPool,
                                         pool_recycle=600
                                         ))

    SESSION.clusto_version = None

    if config.has_option('clusto', 'versioning'):
        SESSION.clusto_versioning_enabled = config.getboolean('clusto', 'versioning')
    else:
        SESSION.clusto_versioning_enabled = True

    try:
        memcache_servers = config.get('clusto', 'memcached').split(',')
#       Memcache should only be imported if we're actually using it, yes?
        import memcache
        logging.info('Memcache server list: %s' % config.get('clusto', 'memcached'))
        SESSION.memcache = memcache.Client(memcache_servers, debug=0)
    except:
        SESSION.memcache = None


def checkDBcompatibility(dbver):

    if dbver == VERSION:
        return True


init_semaphore = threading.Semaphore()
def init_clusto():
    """Initialize a clusto database. """

    init_semaphore.acquire()
    try:
        METADATA.create_all(SESSION.bind)
        c = ClustoMeta()
        flush()
    except Exception, x:
        raise x
    finally:
        init_semaphore.release()


def flush():
    """Flush changes made to clusto objects to the database."""

    SESSION.flush()


def clear():
    """Clear the changes made to objects in the current session. """

    SESSION.expunge_all()
    if hasattr(SESSION(), "TRANSACTIONCOUNTER"):
      del SESSION().TRANSACTIONCOUNTER


def get_driver_name(name):
    "Return driver name given a name, Driver class, or Driver/Entity instance."

    if isinstance(name, basestring):
        if name in DRIVERLIST:
            return name
        else:
            raise NameError("driver name %s doesn't exist." % name)
    elif isinstance(name, type):
        return name._driver_name

    elif isinstance(name, Entity):
        return name.driver
    else:
        raise LookupError("Couldn't find driver name.")

def get_type_name(name):

    if isinstance(name, basestring):
        if name in TYPELIST:
            return name
        else:
            raise NameError("driver name %s doesn't exist." % name)

    elif isinstance(name, type):
        return name._clusto_type
    elif isinstance(name, Entity):
        return name.type
    else:
        raise LookupError("Couldn't find type name.")


def get_driver(entity):
    """Return the driver to use for a given entity """

    if entity.driver in DRIVERLIST:
        return DRIVERLIST[entity.driver]

    return Driver

def get_entities(names=(), clusto_types=(), clusto_drivers=(), attrs=()):
    """Get entities matching the given criteria

    parameters:
      names - list of strings; names to match
      clustotypes - list of strings or Drivers; clustotypes to match
      clustodrivers - list of strings or Drivers; clustodrivers to get
      attrs - list of dicts with the following valid keys: key, number, subkey,
              value ; attribute parameters
    """

    query = Entity.query()

    if names:
        names = [ u'%s' % _ for _ in names ]
        query = query.filter(Entity.name.in_(names))

    if clusto_types:
        ctl = [get_type_name(n) for n in clusto_types]
        query = query.filter(Entity.type.in_(ctl))

    if clusto_drivers:
        cdl = [get_driver_name(n) for n in clusto_drivers]
        query = query.filter(Entity.driver.in_(cdl))

    if attrs:
        for k,v in attrs[0].items():
            if not isinstance(v,str):
                continue
            attrs[0][k] = u'%s' % v
        query = query.filter(Attribute.entity_id==Entity.entity_id)
        query = query.filter(or_(*[Attribute.queryarg(**args)
                                   for args in attrs]))

    return [Driver(entity) for entity in query.all()]


def get_from_pools(pools, clusto_types=(), clusto_drivers=(), search_children=True):
    """Get entitis that are in all the given pools

    parameters:
      pools - list of Pools or strings; pools you want the intersection for
      clusto_types - list of Drivers or strings; clusto types to filter on
      clusto_drivers - list of Drivers or strings; clusto drivers to filter on
    """

    pool_names = []
    pool_types = []

    for p in pools:
        if isinstance(p, basestring):
            pool_names.append(u'%s' % p)
        elif isinstance(p, drivers.Pool):
            pool_types.append(p)
        else:
            raise TypeError("%s is neither a string or a Pool." % str(p))

    pls = pool_types
    for pool_name in pool_names:
        pls.append(get_by_name(pool_name))

    resultsets = []
    for p in pls:
        contents = set(p.contents(clusto_types=clusto_types,
                                  clusto_drivers=clusto_drivers,
                                  search_children=search_children))
        resultsets.append(contents)

    return reduce(set.intersection, resultsets)

def get_by_name(name, assert_driver=None):
    """Return the entity with the given name.

    parameters:
      name - string; the name of the entity
      assert_driver - class; driver class you want to assert this object as
    """

    name = u'%s' % name

    try:
        entity = Entity.query().filter_by(name=name).one()

        retval = Driver(entity)

        if assert_driver:
            if not isinstance(retval, assert_driver):
                raise TypeError("The object %s is not an instance of %s" % (name, assert_driver))

        return retval
    except InvalidRequestError:
        raise LookupError(name + " does not exist.")


def get_by_names(names):
    """Return a list of entities matching the given list of names.

    This will return the entities in the same order as the passed argument,
    placing None in the slot with names that don't exist.

    parameters:
      name - list of strings; names of the entities
    """

    unicode_names = [unicode(name) for name in names]

    entities = Entity.query().filter(Entity.name.in_(unicode_names)).all()

    retvals = [None for x in xrange(len(names))]

    for entity in entities:
        try:
            retvals[names.index(entity.name)] = Driver(entity)
        except:
            pass

    return retvals


get_by_attr = drivers.base.Driver.get_by_attr

def get_or_create(name, driver, **kwargs):
    try:
        obj = get_by_name(name)
    except LookupError:
        obj = driver(name, **kwargs)
        logging.info('Created %s' % obj)
    return obj

def get_by_mac(mac):
    return get_entities(attrs=[{
        'subkey': 'mac',
        'value': mac
    }])

def get_by_serial(serial):
    return get_entities(attrs=[{
        'key': 'system',
        'subkey': 'serial',
        'value': serial,
    }])

def get_by_ip(ip):
    return get_entities(attrs=[{
        'key': 'ip',
        'subkey': 'ipstring',
        'value': ip,
    }])

def get(term):
    if not isinstance(term, basestring):
        raise ValueError('get(term) must be a string')

    try:
        return [get_by_name(term)]
    except LookupError:
        pass

    patterns = [
        ('^(?P<serial>P[0-9]{10})$', get_by_serial, 0),
        ('^(?P<ip>([0-9]{1,3}(\.|$)){4})$', IPManager.get_devices, 0),
        ('^(?P<mac>([0-9a-f]{1,2}([:-]|$)){6})$', get_by_mac, re.IGNORECASE),
    ]

    for pattern, method, flags in patterns:
        match = re.match(pattern, term, flags)
        if match:
            try:
                return method(**match.groupdict())
            except:
                pass
    return None

def rename(oldname, newname):
    """Rename an Entity from oldname to newname.

    THIS CAN CAUSE PROBLEMS IF NOT USED CAREFULLY AND IN ISOLATION FROM OTHER
    ACTIONS.
    """

    old = get_by_name(oldname)

    try:
        begin_transaction()

        new = get_driver(old.entity)(newname)

        for attr in old.attrs(ignore_hidden=False):
            new.add_attr(key=attr.key,
                         number=attr.number,
                         subkey=attr.subkey,
                         value=attr.value)


        for ref in old.references(ignore_hidden=False):
            drivers.base.Driver(ref.entity).add_attr(key=ref.key,
                                                     number=ref.number,
                                                     subkey=ref.subkey,
                                                     value=new)
            ref.delete()

        for counter in SESSION.query(Counter).filter(Counter.entity==old.entity):
            counter.entity = new.entity

        old.entity.delete()
        commit()
    except Exception, x:
        rollback_transaction()
        raise x

def get_latest_version_number():
    "Return the latest version number"

    s = SESSION()

    val = s.execute(latest_version()).fetchone()[0]
    return val


def _check_transaction_counter(not_zero=False):
    tl = SESSION()

    if not hasattr(tl, 'TRANSACTIONCOUNTER'):
        raise TransactionException("No transaction counter.  Outside of a transaction.")

    if tl.TRANSACTIONCOUNTER < 0 or (tl.TRANSACTIONCOUNTER == 0 and not_zero):
        raise TransactionException("Negative transaction counter!  SHOULD NEVER HAPPEN!")

def _init_transaction_counter():

    tl = SESSION()
    if not hasattr(tl, 'TRANSACTIONCOUNTER'):
        tl.TRANSACTIONCOUNTER = 0
    elif tl.TRANSACTIONCOUNTER != 0:
        raise TransactionException("Transaction counter already initialized. At %d." % tl.TRANSACTIONCOUNTER)

def _inc_transaction_counter():
    _check_transaction_counter()

    tl = SESSION()

    tl.TRANSACTIONCOUNTER += 1

def _dec_transaction_counter():

    _check_transaction_counter()

    tl = SESSION()

    tl.TRANSACTIONCOUNTER -= 1


def begin_transaction():
    """Start a transaction

    If already in a transaction start a savepoint transaction.

    If allow_nested is False then an exception will be raised if we're already
    in a transaction.
    """

    if SESSION().is_active:
        _inc_transaction_counter()
        return None
    else:
        _init_transaction_counter()
        _inc_transaction_counter()
        return SESSION.begin()

def rollback_transaction():
    """Rollback a transaction"""

    tl = SESSION()

    try:
      _check_transaction_counter(not_zero=True)
    except TransactionException, x:
      raise TransactionException("Can't rollback, no transaction started.")

    if SESSION.is_active:
        SESSION.rollback()
        _dec_transaction_counter()
    else:
        _dec_transaction_counter()


def change_driver(thingname, newdriver):
    """Change the Driver of a given Entity

    parameters:
      thingname - string; the name of the entity
      newdriver - Driver; the new driver to set for the entity
    """
    if not issubclass(newdriver, Driver):
        raise DriverException("%s is not a driver" % str(newdriver))

    try:
        begin_transaction()

        thing = get_by_name(thingname)

        thing.entity._set_driver_and_type(newdriver._driver_name, newdriver._clusto_type)

        commit()
    except Exception, x:
        rollback_transaction()
        raise x


def commit():
    """Commit changes to the datastore"""

    _check_transaction_counter()
    tl = SESSION()
    if SESSION.is_active:
        if tl.TRANSACTIONCOUNTER == 1:
            exc = None
            for i in range(3): ## retry 3 times
                try:
                    SESSION.commit()
                    break
                except OperationalError, x:
                    exc = x
                    if x.orig[0] == 1213:
                        time.sleep(0.001*i)
                        continue
                    else:
                        raise x
                except ClustoEmptyCommit, x:
                    rollback_transaction()
                    return
            else:
                raise exc
        _dec_transaction_counter()
        flush()


def disconnect():
    SESSION.close()

def delete_entity(entity):
    """Delete an entity and all it's attributes and references"""
    try:
        begin_transaction()
        entity.delete()
        commit()
    except Exception, x:
        rollback_transaction()
        raise x




import clusto
from clusto.test import testbase

from clusto.drivers import IPManager, BasicServer, ResourceTypeException, ResourceException

import IPy

class IPManagerTest(testbase.ClustoTestBase):

    def data(self):

        ip1 = IPManager('a1', gateway='192.168.1.1', netmask='255.255.255.0',
                        baseip='192.168.1.0')

        ip2 = IPManager('b1', gateway='10.0.128.1', netmask='255.255.252.0',
                        baseip='10.0.128.0')

        s = BasicServer('s1')

    def testBadIPAllocation(self):
        
        ip1, ip2, s1 = map(clusto.get_by_name, ['a1', 'b1', 's1'])

        self.assertRaises(ResourceTypeException, ip1.allocate, s1, '10.2.3.4')

    def testNewIP(self):
        
        ip1, ip2, s1 = map(clusto.get_by_name, ['a1', 'b1', 's1'])

        num = 50
        for i in range(num):
            ip1.allocate(s1)


        self.assertEqual(ip1.count, num)
        self.assertEqual(len(ip1.resources(s1)), num)
        
        self.assertEqual(ip1.owners('192.168.1.' + str(num+1)), [s1])

    def testGetIPManager(self):

        ip1, ip2 = map(clusto.get_by_name, ['a1', 'b1'])

        self.assertEqual(ip1, IPManager.get_ip_manager('192.168.1.23'))
        self.assertEqual(ip2, IPManager.get_ip_manager('10.0.129.22'))

    def testGetIP(self):

        ip1, ip2, s1 = map(clusto.get_by_name, ['a1', 'b1', 's1'])

        ip1.allocate(s1)
        ip2.allocate(s1)

        self.assertEqual(sorted(IPManager.get_ips(s1)),
                         sorted(['192.168.1.2', '10.0.128.2']))

    def testReserveIP(self):
        
        ip1, ip2, s1 = map(clusto.get_by_name, ['a1', 'b1', 's1'])

        ip2.allocate(ip2, '10.0.128.4')

        self.assertRaises(ResourceException, ip2.allocate, s1, '10.0.128.4')

        



from clusto.drivers.devices.common import *
from clusto.drivers.devices.servers import *
from clusto.drivers.devices.powerstrips import *
from clusto.drivers.devices.consoleservers import *
from clusto.drivers.devices.networkswitches import *
from clusto.drivers.devices.appliance import *



from clusto.drivers.base import Driver
import sys

class Device(Driver):

    _properties = {'model':None,
                   'serialnum':None,
                   'manufacturer':None}

    _clustotype = "device"
    _driver_name = "device"


    @classmethod
    def get_by_serial_number(self, serialnum):
        pass

    def _get_hostname(self):
        """return a hostname set for this device or its entity name"""

        hostname = self.attrs("hostname")

        if hostname:
            return hostname[0].value
        else:
            return self.entity.name

    def _set_hostname(self, name):

        self.set_attr("hostname", value=name)

    hostname = property(_get_hostname, _set_hostname)

    @property
    def fqdns(self):
        """return the fully qualified domain names for this device"""

        return self.attr_values("fqdn")


    def add_fqdn(self, fqdn):
        """add a fully qualified domain name"""

        if not self.has_attr("fqdn", number=True, value=fqdn):
            self.add_attr("fqdn", number=True, value=fqdn)

    def remove_fqdn(self, fqdn):
        """remove a fully qualified domain name"""

        self.del_attrs("fqdn", number=True, value=fqdn)

    def _power_captcha(self, action='reboot'):
        while True:
            sys.stdout.write('Are you sure you want to %s %s (yes/no)? ' %
                (action, self.name,))
            line = sys.stdin.readline().rstrip('\r\n')
            if line == 'yes':
                return True
            if line == 'no':
                return False
            sys.stdout.write('"yes" or "no", please\n')

    def power_on(self, captcha=True):
        if captcha and not self._power_captcha('power on'):
            return

        ports_set = 0
        for porttype, ports in self.port_info.items():
            if not porttype.startswith('pwr-'): continue
            for portnum, port in ports.items():
                if not port['connection']: continue
                port['connection'].set_power_on(porttype, port['otherportnum'])
                ports_set += 1
        return ports_set

    def power_off(self, captcha=True):
        if captcha and not self._power_captcha('power off'):
            return

        ports_set = 0
        for porttype, ports in self.port_info.items():
            if not porttype.startswith('pwr-'): continue
            for portnum, port in ports.items():
                if not port['connection']: continue
                port['connection'].set_power_off(porttype, port['otherportnum'])
                ports_set += 1
        return ports_set

    def power_reboot(self, captcha=True):
        if captcha and not self._power_captcha('reboot'):
            return

        ports_rebooted = 0
        for porttype, ports in self.port_info.items():
            if not porttype.startswith('pwr-'): continue
            for portnum, port in ports.items():
                if not port['connection']: continue
                port['connection'].reboot(porttype, port['otherportnum'])
                ports_rebooted += 1
        return ports_rebooted

    def console(self, ssh_user='root'):
        console = self.port_info['console-serial'][1]
        if not console['connection']:
            sys.stderr.write('No console connected to %s console-serial:1\n' % self.name)
            sys.stderr.flush()
            return

        if not hasattr(console['connection'], 'console'):
            sys.stderr.write('No console method on %s\n' % console.name)
            sys.stderr.flush()
            return

        console['connection'].connect('console-serial', console['otherportnum'], ssh_user)



#!/usr/bin/env python
# -*- mode: python; sh-basic-offset: 4; indent-tabs-mode: nil; coding: utf-8 -*-
# vim: tabstop=4 softtabstop=4 expandtab shiftwidth=4 fileencoding=utf-8

import argparse
import sys

import clusto
from clusto import drivers
from clusto import script_helper


class ListAll(script_helper.Script):
    '''
    Lists all entities of a type
    '''

    def __init__(self):
        script_helper.Script.__init__(self)

    def _add_arguments(self, parser):
        parser.add_argument('--type', default=None, action='store',
            dest='clusto_type', help='Type of entities to show')

    def run(self, args):
        if args.clusto_type is None:
            clusto_types = []
        else:
            clusto_types = [args.clusto_type]
        for entity in clusto.get_entities(clusto_types=clusto_types):
            print entity.name


def main():
    lp, args = script_helper.init_arguments(ListAll)
    return(lp.run(args))

if __name__ == '__main__':
    sys.exit(main())





from clusto.drivers.base import Device
from clusto.drivers.devices.common import PortMixin, IPMixin

class BasicPowerStrip(IPMixin, PortMixin, Device):
    """
    Basic power strip Driver.
    """

    _clusto_type = "powerstrip"
    _driver_name = "basicpowerstrip"
    

    
    _portmeta = { 'pwr-nema-5' : { 'numports':24, }, 
                  'nic-eth': { 'numports':1, },
                  }
